"""Test ErrorOccurrence module."""

import datetime
from typing import List

from opentrons.protocol_engine.errors.error_occurrence import ErrorOccurrence


def test_error_occurrence_schema() -> None:
    """Test that the schema is overwritten succesfully.

    This is explicitly tested because we are overriding the schema
    due to a default value for errorCode.
    """
    required_items: List[str] = ErrorOccurrence.model_json_schema()["$defs"][
        "ErrorOccurrence"
    ]["required"]
    assert "errorCode" in required_items


def test_parse_error_occurrence() -> None:
    """Test that parsing succeeds without an errorCode.

    This is explicitly tested to ensure backwards compatability is maintained.
    """
    input = '{"id": "abcdefg","errorType": "a bad one","createdAt": "2023-06-12 15:08:54.730451","detail": "This is a bad error"}'

    result = ErrorOccurrence.model_validate_json(input)

    expected = ErrorOccurrence(
        id="abcdefg",
        errorType="a bad one",
        createdAt=datetime.datetime.fromisoformat("2023-06-12 15:08:54.730451"),
        detail="This is a bad error",
        errorCode="4000",
    )

    assert result == expected



from typing import cast

from opentrons import protocol_api, types

metadata = {
    "protocolName": "Mad Mag",
    "author": "Opentrons <engineering@opentrons.com>",
    "description": "A Magnetic Module Test",
    "source": "Opentrons Repository",
    "apiLevel": "2.2",
}


def run(ctx: protocol_api.ProtocolContext) -> None:
    ctx.home()
    tr = ctx.load_labware("opentrons_96_tiprack_300ul", 1)
    mm = cast(protocol_api.MagneticModuleContext, ctx.load_module("magnetic module", 4))
    lw = mm.load_labware("nest_96_wellplate_100ul_pcr_full_skirt")
    right = ctx.load_instrument("p300_single_gen2", types.Mount.RIGHT, [tr])

    mm.disengage()

    right.transfer(30, lw.wells()[0].bottom(1), lw.wells()[1].bottom(1))

    mm.engage()
    mm.disengage()

    mm.engage(height=30)
    mm.disengage()

    mm.engage(offset=-10)
    mm.disengage()

    mm.engage(height_from_base=15)
    mm.disengage()



"""Liquid state store tests.

DEPRECATED: Testing LiquidClassStore independently of LiquidClassView is no
longer helpful. Try to add new tests to test_liquid_class_state.py, where they can be
tested together, treating LiquidClassState as a private implementation detail.
"""

import pytest

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    LiquidClassSchemaV1,
)

from .command_fixtures import create_comment_command
from opentrons.protocol_engine import actions
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.liquid_classes import LiquidClassStore
from opentrons.protocol_engine.types import LiquidClassRecord


@pytest.fixture
def subject() -> LiquidClassStore:
    """The LiquidClassStore test subject."""
    return LiquidClassStore()


def test_handles_add_liquid_class(
    subject: LiquidClassStore, minimal_liquid_class_def2: LiquidClassSchemaV1
) -> None:
    """Should add the LiquidClassRecord to the store."""
    pipette_0 = minimal_liquid_class_def2.byPipette[0]
    by_tip_type_0 = pipette_0.byTipType[0]
    liquid_class_record = LiquidClassRecord(
        liquidClassName=minimal_liquid_class_def2.liquidClassName,
        pipetteModel=pipette_0.pipetteModel,
        tiprack=by_tip_type_0.tiprack,
        aspirate=by_tip_type_0.aspirate,
        singleDispense=by_tip_type_0.singleDispense,
        multiDispense=by_tip_type_0.multiDispense,
    )

    subject.handle_action(
        actions.SucceedCommandAction(
            command=create_comment_command(),
            state_update=update_types.StateUpdate(
                liquid_class_loaded=update_types.LiquidClassLoadedUpdate(
                    liquid_class_id="liquid-class-id",
                    liquid_class_record=liquid_class_record,
                ),
            ),
        )
    )

    assert len(subject.state.liquid_class_record_by_id) == 1
    assert (
        subject.state.liquid_class_record_by_id["liquid-class-id"]
        == liquid_class_record
    )

    assert len(subject.state.liquid_class_record_to_id) == 1
    # Make sure that LiquidClassRecords are hashable, and that we can query for LiquidClassRecords by value:
    assert (
        subject.state.liquid_class_record_to_id[liquid_class_record]
        == "liquid-class-id"
    )
    # If this fails with an error like "TypeError: unhashable type: AspirateProperties", then you broke something.



# Uncomment to enable logging during tests
from __future__ import annotations

import asyncio
import contextlib
import inspect
import io
import json
import os
import pathlib
import zipfile
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Callable,
    Dict,
    Generator,
    NamedTuple,
    TextIO,
    Union,
    cast,
)

import pytest
from _pytest.fixtures import SubRequest
from decoy import Decoy
from typing_extensions import TypedDict

from opentrons_shared_data.liquid_classes.types import (
    AspiratePropertiesDict,
    BlowoutPropertiesDict,
    DelayPropertiesDict,
    MixPropertiesDict,
    RetractAspirateDict,
    RetractDispenseDict,
    SingleDispensePropertiesDict,
    SubmergeDict,
    TipPositionDict,
    TouchTipPropertiesDict,
    TransferPropertiesDict,
)

from opentrons.protocol_engine.types import PostRunHardwareState

try:
    import aionotify  # type: ignore[import-untyped]
except (OSError, ModuleNotFoundError):
    aionotify = None

from opentrons_shared_data.deck import (
    DEFAULT_DECK_DEFINITION_VERSION,
)
from opentrons_shared_data.deck import (
    load as load_deck,
)
from opentrons_shared_data.deck.types import (
    DeckDefinitionV3,
    DeckDefinitionV5,
    RobotModel,
)
from opentrons_shared_data.labware.types import LabwareDefinition, LabwareDefinition2
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    AspirateProperties,
    BlowoutLocation,
    BlowoutParams,
    BlowoutProperties,
    ByPipetteSetting,
    ByTipTypeSetting,
    Coordinate,
    DelayParams,
    DelayProperties,
    LiquidClassSchemaV1,
    LiquidClassTouchTipParams,
    MixParams,
    MixProperties,
    MultiDispenseProperties,
    PositionReference,
    RetractAspirate,
    RetractDispense,
    SingleDispenseProperties,
    Submerge,
    TipPosition,
    TouchTipProperties,
)
from opentrons_shared_data.module.types import ModuleDefinitionV3
from opentrons_shared_data.protocol.types import JsonProtocol
from opentrons_shared_data.robot.types import RobotTypeEnum

from opentrons import config
from opentrons.drivers.rpi_drivers.gpio_simulator import SimulatingGPIOCharDev
from opentrons.hardware_control import (
    API,
    HardwareControlAPI,
    ThreadManagedHardware,
    ThreadManager,
)
from opentrons.protocol_api import Labware, ProtocolContext, create_protocol_context
from opentrons.protocol_api.core.legacy.deck import (
    DEFAULT_LEGACY_DECK_DEFINITION_VERSION,
)
from opentrons.protocol_api.core.legacy.legacy_labware_core import LegacyLabwareCore
from opentrons.protocol_engine import (
    Config as ProtocolEngineConfig,
)
from opentrons.protocol_engine import (
    DeckType,
    error_recovery_policy,
)
from opentrons.protocol_engine.create_protocol_engine import (
    create_protocol_engine_in_thread,
)
from opentrons.protocols.api_support import deck_type
from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION
from opentrons.protocols.api_support.types import APIVersion
from opentrons.types import Location, Point

if TYPE_CHECKING:
    from opentrons.drivers.smoothie_drivers import SmoothieDriver as SmoothieDriverType


class Protocol(NamedTuple):
    text: str
    filename: str
    filelike: TextIO


class Bundle(TypedDict):
    source_dir: pathlib.Path
    filename: str
    contents: str
    filelike: io.BytesIO
    binary_zipfile: bytes
    metadata: Dict[str, str]
    bundled_data: Dict[str, bytes]
    bundled_labware: Dict[str, LabwareDefinition]
    bundled_python: Dict[str, Any]


@pytest.fixture()
def ot_config_tempdir(tmp_path: pathlib.Path) -> Generator[pathlib.Path, None, None]:
    os.environ["OT_API_CONFIG_DIR"] = str(tmp_path)
    config.reload()

    yield tmp_path

    del os.environ["OT_API_CONFIG_DIR"]
    config.reload()


@pytest.fixture()
def is_robot(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(config, "IS_ROBOT", True)


@pytest.fixture
def mock_feature_flags(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    for name, func in inspect.getmembers(config.feature_flags, inspect.isfunction):
        params = inspect.getfullargspec(func)
        mock_get_ff = decoy.mock(func=func)
        if any("robot_type" in p for p in params.args):
            decoy.when(mock_get_ff(RobotTypeEnum.FLEX)).then_return(False)
        else:
            decoy.when(mock_get_ff()).then_return(False)
        monkeypatch.setattr(config.feature_flags, name, mock_get_ff)


@pytest.fixture
async def enable_ot3_hardware_controller(
    request: pytest.FixtureRequest,
    decoy: Decoy,
    mock_feature_flags: None,
) -> None:
    # this is from the command line parameters added in root conftest
    if request.config.getoption("--ot2-only"):
        pytest.skip("testing only ot2")

    decoy.when(config.feature_flags.enable_ot3_hardware_controller()).then_return(True)


@pytest.fixture()
def protocol_file() -> str:
    return "testosaur_v2.py"


@pytest.fixture()
def protocol(protocol_file: str) -> Generator[Protocol, None, None]:
    filename = os.path.join(os.path.dirname(__file__), "data", protocol_file)
    with open(filename, encoding="utf-8") as file:
        text = file.read()
        file.seek(0)
        yield Protocol(text=text, filename=filename, filelike=file)


@pytest.fixture()
def virtual_smoothie_env(monkeypatch: pytest.MonkeyPatch) -> None:
    # TODO (ben 20180426): move this to the .env file
    monkeypatch.setenv("ENABLE_VIRTUAL_SMOOTHIE", "true")


@pytest.fixture(params=["ot2", "ot3"])
async def machine_variant_ffs(
    request: SubRequest,
    decoy: Decoy,
    mock_feature_flags: None,
) -> None:
    device_param = request.param

    if request.node.get_closest_marker("ot3_only") and device_param == "ot2":
        pytest.skip()
    if request.node.get_closest_marker("ot2_only") and device_param == "ot3":
        pytest.skip()

    decoy.when(config.feature_flags.enable_ot3_hardware_controller()).then_return(
        device_param == "ot3"
    )


@contextlib.asynccontextmanager
async def _build_ot2_hw() -> AsyncGenerator[ThreadManagedHardware, None]:
    hw_sim = ThreadManager(API.build_hardware_simulator)
    old_config = config.robot_configs.load()
    try:
        yield hw_sim
    finally:
        config.robot_configs.clear()
        for m in hw_sim.wrapped().attached_modules:
            await m.cleanup()
        hw_sim.set_config(old_config)
        hw_sim.clean_up()


@pytest.fixture()
async def ot2_hardware(
    virtual_smoothie_env: None,
) -> AsyncGenerator[ThreadManagedHardware, None]:
    async with _build_ot2_hw() as hw:
        yield hw


@contextlib.asynccontextmanager
async def _build_ot3_hw() -> AsyncGenerator[ThreadManagedHardware, None]:
    from opentrons.hardware_control.ot3api import OT3API

    hw_sim = ThreadManager(OT3API.build_hardware_simulator)
    old_config = config.robot_configs.load()
    try:
        yield hw_sim
    finally:
        config.robot_configs.clear()
        for m in hw_sim.wrapped().attached_modules:
            await m.cleanup()
        hw_sim.set_config(old_config)
        hw_sim.clean_up()


@pytest.fixture()
async def ot3_hardware(
    request: pytest.FixtureRequest,
    enable_ot3_hardware_controller: None,
) -> AsyncGenerator[ThreadManagedHardware, None]:
    # this is from the command line parameters added in root conftest
    if request.config.getoption("--ot2-only"):
        pytest.skip("testing only ot2")
    async with _build_ot3_hw() as hw:
        yield hw


@pytest.fixture(params=["OT-2 Standard", "OT-3 Standard"])
async def robot_model(
    request: pytest.FixtureRequest,
    decoy: Decoy,
    mock_feature_flags: None,
    virtual_smoothie_env: None,
) -> AsyncGenerator[RobotModel, None]:
    which_machine = cast(RobotModel, request.param)
    if request.node.get_closest_marker("ot2_only") and which_machine == "OT-3 Standard":
        pytest.skip("test requests only ot-2")
    if request.node.get_closest_marker("ot3_only") and which_machine == "OT-2 Standard":
        pytest.skip("test requests only ot-3")
    if which_machine == "OT-3 Standard" and request.config.getoption("--ot2-only"):
        pytest.skip("testing only ot2")

    decoy.when(config.feature_flags.enable_ot3_hardware_controller()).then_return(
        which_machine == "OT-3 Standard"
    )
    yield which_machine


@pytest.fixture
def deck_definition_name(robot_model: RobotModel) -> str:
    if robot_model == "OT-3 Standard":
        return deck_type.STANDARD_OT3_DECK
    elif robot_model == "OT-2 Standard":
        # There are two OT-2 deck definitions (standard and short-trash),
        # but RobotModel does not draw such a distinction. We assume here that it's
        # sufficient to run OT-2 tests with the standard deck definition only.
        return deck_type.STANDARD_OT2_DECK


@pytest.fixture
def deck_definition(deck_definition_name: str) -> DeckDefinitionV5:
    return load_deck(deck_definition_name, DEFAULT_DECK_DEFINITION_VERSION)


@pytest.fixture
def legacy_deck_definition(deck_definition_name: str) -> DeckDefinitionV3:
    return load_deck(deck_definition_name, DEFAULT_LEGACY_DECK_DEFINITION_VERSION)


@pytest.fixture()
async def hardware(
    request: SubRequest,
    decoy: Decoy,
    mock_feature_flags: None,
    virtual_smoothie_env: None,
    robot_model: RobotModel,
) -> AsyncGenerator[ThreadManagedHardware, None]:
    hw_builder = {"OT-2 Standard": _build_ot2_hw, "OT-3 Standard": _build_ot3_hw}[
        robot_model
    ]

    async with hw_builder() as hw:
        decoy.when(config.feature_flags.enable_ot3_hardware_controller()).then_return(
            robot_model == "OT-3 Standard"
        )

        yield hw


def _make_ot2_non_pe_ctx(
    hardware: ThreadManagedHardware, deck_type: str
) -> ProtocolContext:
    """Return a ProtocolContext configured for an OT-2 and not backed by Protocol Engine."""
    return create_protocol_context(
        api_version=APIVersion(2, 13), hardware_api=hardware, deck_type=deck_type
    )


@contextlib.contextmanager
def _make_ot3_pe_ctx(
    hardware: ThreadManagedHardware,
    deck_type: str,
) -> Generator[ProtocolContext, None, None]:
    """Return a ProtocolContext configured for an OT-3 and backed by Protocol Engine."""
    with create_protocol_engine_in_thread(
        hardware_api=hardware.wrapped(),
        config=ProtocolEngineConfig(
            robot_type="OT-3 Standard",
            deck_type=DeckType.OT3_STANDARD,
            ignore_pause=True,
            use_virtual_pipettes=True,
            use_virtual_modules=True,
            use_virtual_gripper=True,
            # TODO figure out if we will want to use a "real" deck config here or if we are fine with simulated
            use_simulated_deck_config=True,
            block_on_door_open=False,
        ),
        deck_configuration=None,
        file_provider=None,
        error_recovery_policy=error_recovery_policy.never_recover,
        drop_tips_after_run=False,
        post_run_hardware_state=PostRunHardwareState.STAY_ENGAGED_IN_PLACE,
        # TODO(jbl 10-30-2023) load_fixed_trash being hardcoded to True will be refactored once we need tests to have
        #   this be False
        load_fixed_trash=True,
        camera_provider=None,
    ) as (
        engine,
        loop,
    ):
        yield create_protocol_context(
            api_version=MAX_SUPPORTED_VERSION,
            hardware_api=hardware,
            deck_type=deck_type,
            protocol_engine=engine,
            protocol_engine_loop=loop,
        )


@pytest.fixture()
def ctx(
    request: pytest.FixtureRequest,
    robot_model: RobotModel,
    hardware: ThreadManagedHardware,
    deck_definition_name: str,
) -> Generator[ProtocolContext, None, None]:
    if robot_model == "OT-2 Standard":
        yield _make_ot2_non_pe_ctx(hardware=hardware, deck_type=deck_definition_name)
    elif robot_model == "OT-3 Standard":
        if request.node.get_closest_marker("apiv2_non_pe_only"):
            pytest.skip("Test requests only non-Protocol-Engine ProtocolContexts")
        else:
            with _make_ot3_pe_ctx(
                hardware=hardware, deck_type=deck_definition_name
            ) as ctx:
                yield ctx


@pytest.fixture()
async def smoothie(
    virtual_smoothie_env: None,
    monkeypatch: pytest.MonkeyPatch,
) -> AsyncGenerator[SmoothieDriverType, None]:
    from opentrons.config import robot_configs
    from opentrons.drivers.smoothie_drivers import SmoothieDriver

    driver = SmoothieDriver(
        robot_configs.load_ot2(), SimulatingGPIOCharDev("simulated")
    )
    await driver.connect()
    yield driver
    try:
        await driver.disconnect()
    except AttributeError:
        # if the test disconnected
        pass


@pytest.fixture
def hardware_controller_lockfile(
    monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path
) -> pathlib.Path:
    lockfile_dir = tmp_path / "hardware_controller_lockfile"
    lockfile_dir.mkdir()
    lockfile = lockfile_dir / "hardware.lock"

    monkeypatch.setitem(config.CONFIG, "hardware_controller_lockfile", lockfile)

    return lockfile_dir


@pytest.fixture()
async def hardware_api(is_robot: None) -> HardwareControlAPI:
    hw_api = await API.build_hardware_simulator(loop=asyncio.get_running_loop())
    return hw_api


@pytest.fixture()
def get_labware_fixture() -> Callable[[str], LabwareDefinition]:
    def _get_labware_fixture(fixture_name: str) -> LabwareDefinition:
        with open(
            (
                pathlib.Path(__file__).parent
                / ".."
                / ".."
                / ".."
                / "shared-data"
                / "labware"
                / "fixtures"
                / "2"
                / f"{fixture_name}.json"
            ),
            "rb",
        ) as f:
            return cast(LabwareDefinition, json.loads(f.read().decode("utf-8")))

    return _get_labware_fixture


@pytest.fixture()
def get_json_protocol_fixture() -> Callable[[str, str, bool], Union[str, JsonProtocol]]:
    def _get_json_protocol_fixture(
        fixture_version: str,
        fixture_name: str,
        decode: bool = True,
    ) -> Union[str, JsonProtocol]:
        with open(
            pathlib.Path(__file__).parent
            / ".."
            / ".."
            / ".."
            / "shared-data"
            / "protocol"
            / "fixtures"
            / fixture_version
            / f"{fixture_name}.json",
            "rb",
        ) as f:
            contents = f.read().decode("utf-8")
            if decode:
                return cast(JsonProtocol, json.loads(contents))
            else:
                return contents

    return _get_json_protocol_fixture


@pytest.fixture
def get_bundle_fixture() -> Callable[[str], Bundle]:
    def get_std_labware(loadName: str, version: int = 1) -> LabwareDefinition:
        with open(
            pathlib.Path(__file__).parent
            / ".."
            / ".."
            / ".."
            / "shared-data"
            / "labware"
            / "definitions"
            / "2"
            / loadName
            / f"{version}.json",
            "rb",
        ) as f:
            labware_def = cast(LabwareDefinition, json.loads(f.read().decode("utf-8")))
        return labware_def

    def _get_bundle_protocol_fixture(fixture_name: str) -> Bundle:
        """
        It's ugly to store bundles as .zip's, so we'll build the .zip
        from fixtures and return it as `bytes`.
        We also need to hard-code fixture data here (bundled_labware,
        bundled_python, bundled_data, metadata) for the tests to use in
        their assertions.
        """
        fixture_dir = (
            pathlib.Path(__file__).parent
            / "protocols"
            / "fixtures"
            / "bundled_protocols"
            / fixture_name
        )

        result: Bundle = {  # type: ignore[typeddict-item]
            "filename": f"{fixture_name}.zip",
            "source_dir": fixture_dir,
        }

        fixed_trash_def = get_std_labware("opentrons_1_trash_1100ml_fixed")

        empty_protocol = "def run(context):\n    pass"

        if fixture_name == "simple_bundle":
            with open(fixture_dir / "protocol.py", "r") as f:
                result["contents"] = f.read()
            with open(fixture_dir / "data.txt", "rb") as f:
                result["bundled_data"] = {"data.txt": f.read()}
            with open(fixture_dir / "custom_labware.json", "r") as f:
                custom_labware = json.load(f)

            tiprack_def = get_std_labware("opentrons_96_tiprack_10ul")
            result["bundled_labware"] = {
                "opentrons/opentrons_1_trash_1100ml_fixed/1": fixed_trash_def,
                "custom_beta/custom_labware/1": custom_labware,
                "opentrons/opentrons_96_tiprack_10ul/1": tiprack_def,
            }
            result["bundled_python"] = {}

            # NOTE: this is copy-pasted from the .py fixture file
            result["metadata"] = {"author": "MISTER FIXTURE", "apiLevel": "2.0"}

            # make binary zipfile
            binary_zipfile = io.BytesIO()
            with zipfile.ZipFile(binary_zipfile, "w") as z:
                z.writestr("labware/custom_labware.json", json.dumps(custom_labware))
                z.writestr("labware/tiprack.json", json.dumps(tiprack_def))
                z.writestr("labware/fixed_trash.json", json.dumps(fixed_trash_def))
                z.writestr("protocol.ot2.py", result["contents"])
                z.writestr("data/data.txt", result["bundled_data"]["data.txt"])
            binary_zipfile.seek(0)
            result["binary_zipfile"] = binary_zipfile.read()
            binary_zipfile.seek(0)
            result["filelike"] = binary_zipfile

        elif fixture_name == "no_root_files_bundle":
            binary_zipfile = io.BytesIO()
            with zipfile.ZipFile(binary_zipfile, "w") as z:
                z.writestr("inner_dir/protocol.ot2.py", empty_protocol)
            binary_zipfile.seek(0)
            result["binary_zipfile"] = binary_zipfile.read()
            binary_zipfile.seek(0)
            result["filelike"] = binary_zipfile
        elif fixture_name == "no_entrypoint_protocol_bundle":
            binary_zipfile = io.BytesIO()
            with zipfile.ZipFile(binary_zipfile, "w") as z:
                z.writestr("rando_pyfile_name.py", empty_protocol)
            binary_zipfile.seek(0)
            result["binary_zipfile"] = binary_zipfile.read()
            binary_zipfile.seek(0)
            result["filelike"] = binary_zipfile
        elif fixture_name == "conflicting_labware_bundle":
            binary_zipfile = io.BytesIO()
            with zipfile.ZipFile(binary_zipfile, "w") as z:
                plate_def = get_std_labware("biorad_96_wellplate_200ul_pcr")
                z.writestr("protocol.ot2.py", empty_protocol)
                z.writestr("labware/fixed_trash.json", json.dumps(fixed_trash_def))
                z.writestr("labware/plate.json", json.dumps(plate_def))
                z.writestr("labware/same_plate.json", json.dumps(plate_def))
            binary_zipfile.seek(0)
            result["binary_zipfile"] = binary_zipfile.read()
            binary_zipfile.seek(0)
            result["filelike"] = binary_zipfile
        elif fixture_name == "missing_labware_bundle":
            # parsing should fail b/c this bundle lacks labware defs.
            with open(fixture_dir / "protocol.py", "r") as f:
                protocol_contents = f.read()
            binary_zipfile = io.BytesIO()
            with zipfile.ZipFile(binary_zipfile, "w") as z:
                z.writestr("protocol.ot2.py", protocol_contents)
            binary_zipfile.seek(0)
            result["binary_zipfile"] = binary_zipfile.read()
            binary_zipfile.seek(0)
            result["filelike"] = binary_zipfile
        else:
            raise ValueError(
                f'get_bundle_fixture has no case to handle fixture "{fixture_name}"'
            )
        return result

    return _get_bundle_protocol_fixture


@pytest.fixture()
def minimal_labware_def() -> LabwareDefinition2:
    return {
        "metadata": {
            "displayName": "minimal labware",
            "displayCategory": "other",
            "displayVolumeUnits": "mL",
        },
        "allowedRoles": ["labware"],
        "cornerOffsetFromSlot": {"x": 10, "y": 10, "z": 5},
        "parameters": {
            "isTiprack": False,
            "loadName": "minimal_labware_def",
            "isMagneticModuleCompatible": True,
            "format": "irregular",
        },
        "ordering": [["A1"], ["A2"]],
        "wells": {
            "A1": {
                "depth": 40,
                "totalLiquidVolume": 100,
                "diameter": 30,
                "x": 0,
                "y": 0,
                "z": 0,
                "shape": "circular",
            },
            "A2": {
                "depth": 40,
                "totalLiquidVolume": 100,
                "diameter": 30,
                "x": 10,
                "y": 0,
                "z": 0,
                "shape": "circular",
            },
        },
        "dimensions": {"xDimension": 1.0, "yDimension": 2.0, "zDimension": 3.0},
        "groups": [],
        "brand": {"brand": "opentrons"},
        "version": 1,
        "schemaVersion": 2,
        "namespace": "opentronstest",
    }


@pytest.fixture()
def minimal_labware_def2() -> LabwareDefinition2:
    return {
        "metadata": {
            "displayName": "other test labware",
            "displayCategory": "other",
            "displayVolumeUnits": "mL",
        },
        "cornerOffsetFromSlot": {"x": 10, "y": 10, "z": 5},
        "parameters": {
            "isTiprack": False,
            "loadName": "minimal_labware_def",
            "isMagneticModuleCompatible": True,
            "format": "irregular",
        },
        "ordering": [["A1", "B1", "C1"], ["A2", "B2", "C2"]],
        "wells": {
            "A1": {
                "depth": 40,
                "totalLiquidVolume": 100,
                "diameter": 30,
                "x": 0,
                "y": 18,
                "z": 0,
                "shape": "circular",
            },
            "B1": {
                "depth": 40,
                "totalLiquidVolume": 100,
                "diameter": 30,
                "x": 0,
                "y": 9,
                "z": 0,
                "shape": "circular",
            },
            "C1": {
                "depth": 40,
                "totalLiquidVolume": 100,
                "diameter": 30,
                "x": 0,
                "y": 0,
                "z": 0,
                "shape": "circular",
            },
            "A2": {
                "depth": 40,
                "totalLiquidVolume": 100,
                "diameter": 30,
                "x": 9,
                "y": 18,
                "z": 0,
                "shape": "circular",
            },
            "B2": {
                "depth": 40,
                "totalLiquidVolume": 100,
                "diameter": 30,
                "x": 9,
                "y": 9,
                "z": 0,
                "shape": "circular",
            },
            "C2": {
                "depth": 40,
                "totalLiquidVolume": 100,
                "diameter": 30,
                "x": 9,
                "y": 0,
                "z": 0,
                "shape": "circular",
            },
        },
        "groups": [],
        "dimensions": {"xDimension": 1.0, "yDimension": 2.0, "zDimension": 3.0},
        "schemaVersion": 2,
        "version": 1,
        "namespace": "dummy_namespace",
        "brand": {"brand": "opentrons"},
    }


@pytest.fixture()
def min_lw_impl(minimal_labware_def: LabwareDefinition2) -> LegacyLabwareCore:
    return LegacyLabwareCore(
        definition=minimal_labware_def, parent=Location(Point(0, 0, 0), "deck")
    )


@pytest.fixture()
def min_lw2_impl(minimal_labware_def2: LabwareDefinition2) -> LegacyLabwareCore:
    return LegacyLabwareCore(
        definition=minimal_labware_def2, parent=Location(Point(0, 0, 0), "deck")
    )


@pytest.fixture()
def min_lw(min_lw_impl: LegacyLabwareCore) -> Labware:
    return Labware(
        core=min_lw_impl,
        api_version=APIVersion(2, 13),
        protocol_core=None,  # type: ignore[arg-type]
        core_map=None,  # type: ignore[arg-type]
    )


@pytest.fixture()
def min_lw2(min_lw2_impl: LegacyLabwareCore) -> Labware:
    return Labware(
        core=min_lw2_impl,
        api_version=APIVersion(2, 13),
        protocol_core=None,  # type: ignore[arg-type]
        core_map=None,  # type: ignore[arg-type]
    )


@pytest.fixture()
def minimal_module_def() -> ModuleDefinitionV3:
    return {
        "$otSharedSchema": "module/schemas/3",
        "moduleType": "temperatureModuleType",
        "model": "temperatureModuleV1",
        "labwareOffset": {"x": -0.15, "y": -0.15, "z": 80.09},
        "features": {},
        "extents": {
            "total": {
                "backLeftBottom": {"x": -5, "y": -10, "z": -15},
                "frontRightTop": {"x": 15, "y": -20, "z": 30},
            },
        },
        "dimensions": {
            "bareOverallHeight": 84,
            "overLabwareHeight": 0,
            "xDimension": 123,
            "yDimension": 321,
            "labwareInterfaceXDimension": 123,
            "labwareInterfaceYDimension": 321,
        },
        "calibrationPoint": {"x": 12.0, "y": 8.75, "z": 0.0},
        "config": {},
        "displayName": "Sample Module",
        "quirks": [],
        "slotTransforms": {},
        "orientation": {
            "1": "left",
            "2": "center",
            "3": "right",
            "4": "left",
            "5": "center",
            "6": "right",
            "7": "left",
            "8": "center",
            "9": "right",
            "10": "left",
            "11": "center",
        },
        "compatibleWith": ["temperatureModuleV2"],
        "cornerOffsetFromSlot": {"x": 0.1, "y": 0.1, "z": 0.0},
    }


@pytest.fixture
def minimal_liquid_class_def1() -> LiquidClassSchemaV1:
    return LiquidClassSchemaV1(
        liquidClassName="water1",
        displayName="water 1",
        description="some water",
        schemaVersion=1,
        version=1,
        namespace="test-fixture-1",
        byPipette=[],
    )


@pytest.fixture
def minimal_liquid_class_def2() -> LiquidClassSchemaV1:
    return LiquidClassSchemaV1(
        liquidClassName="water2",
        displayName="water 2",
        description="some water",
        schemaVersion=1,
        version=1,
        namespace="test-fixture-2",
        byPipette=[
            ByPipetteSetting(
                pipetteModel="flex_1channel_50",
                byTipType=[
                    ByTipTypeSetting(
                        tiprack="opentrons_flex_96_tiprack_50ul",
                        aspirate=AspirateProperties(
                            submerge=Submerge(
                                startPosition=TipPosition(
                                    positionReference=PositionReference.WELL_TOP,
                                    offset=Coordinate(x=0, y=0, z=-5),
                                ),
                                speed=100,
                                delay=DelayProperties(
                                    enable=True, params=DelayParams(duration=1.5)
                                ),
                            ),
                            retract=RetractAspirate(
                                endPosition=TipPosition(
                                    positionReference=PositionReference.WELL_TOP,
                                    offset=Coordinate(x=0, y=0, z=5),
                                ),
                                speed=100,
                                airGapByVolume=[(5.0, 3.0), (10.0, 4.0)],
                                touchTip=TouchTipProperties(enable=False),
                                delay=DelayProperties(enable=False),
                            ),
                            aspiratePosition=TipPosition(
                                positionReference=PositionReference.WELL_BOTTOM,
                                offset=Coordinate(x=0, y=0, z=-5),
                            ),
                            flowRateByVolume=[(10.0, 40.0), (20.0, 30.0)],
                            correctionByVolume=[(15.0, 1.5), (30.0, -5.0)],
                            preWet=True,
                            mix=MixProperties(enable=False),
                            delay=DelayProperties(
                                enable=True, params=DelayParams(duration=2)
                            ),
                        ),
                        singleDispense=SingleDispenseProperties(
                            submerge=Submerge(
                                startPosition=TipPosition(
                                    positionReference=PositionReference.LIQUID_MENISCUS,
                                    offset=Coordinate(x=0, y=0, z=-5),
                                ),
                                speed=100,
                                delay=DelayProperties(enable=False),
                            ),
                            retract=RetractDispense(
                                endPosition=TipPosition(
                                    positionReference=PositionReference.WELL_TOP,
                                    offset=Coordinate(x=0, y=0, z=5),
                                ),
                                speed=100,
                                airGapByVolume=[(5.0, 3.0), (10.0, 4.0)],
                                blowout=BlowoutProperties(enable=False),
                                touchTip=TouchTipProperties(enable=False),
                                delay=DelayProperties(enable=False),
                            ),
                            dispensePosition=TipPosition(
                                positionReference=PositionReference.WELL_BOTTOM,
                                offset=Coordinate(x=0, y=0, z=-5),
                            ),
                            flowRateByVolume=[(10.0, 40.0), (20.0, 30.0)],
                            correctionByVolume=[(15.0, -1.5), (30.0, 5.0)],
                            mix=MixProperties(enable=False),
                            pushOutByVolume=[(10.0, 7.0), (20.0, 10.0)],
                            delay=DelayProperties(enable=False),
                        ),
                        multiDispense=None,
                    )
                ],
            )
        ],
    )


@pytest.fixture
def maximal_liquid_class_def() -> LiquidClassSchemaV1:
    """Return a liquid class def with all properties enabled."""
    return LiquidClassSchemaV1(
        liquidClassName="test_water",
        displayName="Test Water",
        description="some water",
        schemaVersion=1,
        version=1,
        namespace="opentrons",
        byPipette=[
            ByPipetteSetting(
                pipetteModel="flex_1channel_50",
                byTipType=[
                    ByTipTypeSetting(
                        tiprack="opentrons_flex_96_tiprack_50ul",
                        aspirate=AspirateProperties(
                            submerge=Submerge(
                                startPosition=TipPosition(
                                    positionReference=PositionReference.WELL_TOP,
                                    offset=Coordinate(x=1, y=2, z=3),
                                ),
                                speed=100,
                                delay=DelayProperties(
                                    enable=True, params=DelayParams(duration=10.0)
                                ),
                            ),
                            retract=RetractAspirate(
                                endPosition=TipPosition(
                                    positionReference=PositionReference.WELL_TOP,
                                    offset=Coordinate(x=3, y=2, z=1),
                                ),
                                speed=50,
                                airGapByVolume=[(1.0, 0.1), (49.9, 0.1), (50.0, 0.0)],
                                touchTip=TouchTipProperties(
                                    enable=True,
                                    params=LiquidClassTouchTipParams(
                                        zOffset=-1, mmFromEdge=0.5, speed=30
                                    ),
                                ),
                                delay=DelayProperties(
                                    enable=True, params=DelayParams(duration=20)
                                ),
                            ),
                            aspiratePosition=TipPosition(
                                positionReference=PositionReference.WELL_BOTTOM,
                                offset=Coordinate(x=10, y=20, z=30),
                            ),
                            flowRateByVolume=[(1.0, 35.0), (10.0, 24.0), (50.0, 35.0)],
                            correctionByVolume=[
                                (0.0, 0.0),
                                (10.0, -1.0),
                                (50.0, -10.0),
                            ],
                            preWet=True,
                            mix=MixProperties(
                                enable=True, params=MixParams(repetitions=1, volume=50)
                            ),
                            delay=DelayProperties(
                                enable=True, params=DelayParams(duration=0.2)
                            ),
                        ),
                        singleDispense=SingleDispenseProperties(
                            submerge=Submerge(
                                startPosition=TipPosition(
                                    positionReference=PositionReference.WELL_TOP,
                                    offset=Coordinate(x=30, y=20, z=10),
                                ),
                                speed=100,
                                delay=DelayProperties(
                                    enable=True, params=DelayParams(duration=1.1)
                                ),
                            ),
                            retract=RetractDispense(
                                endPosition=TipPosition(
                                    positionReference=PositionReference.WELL_TOP,
                                    offset=Coordinate(x=11, y=22, z=33),
                                ),
                                speed=50,
                                airGapByVolume=[(1.0, 0.1), (49.9, 0.1), (50.0, 0.0)],
                                blowout=BlowoutProperties(
                                    enable=True,
                                    params=BlowoutParams(
                                        location=BlowoutLocation.SOURCE,
                                        flowRate=100,
                                    ),
                                ),
                                touchTip=TouchTipProperties(
                                    enable=True,
                                    params=LiquidClassTouchTipParams(
                                        zOffset=-1, mmFromEdge=0.75, speed=30
                                    ),
                                ),
                                delay=DelayProperties(
                                    enable=True, params=DelayParams(duration=10)
                                ),
                            ),
                            dispensePosition=TipPosition(
                                positionReference=PositionReference.WELL_BOTTOM,
                                offset=Coordinate(x=33, y=22, z=11),
                            ),
                            flowRateByVolume=[(1.0, 50.0)],
                            correctionByVolume=[
                                (0.0, 0.0),
                                (10.0, -1.0),
                                (50.0, -10.0),
                            ],
                            mix=MixProperties(
                                enable=True, params=MixParams(repetitions=1, volume=50)
                            ),
                            pushOutByVolume=[
                                (1.0, 7.0),
                                (4.999, 7.0),
                                (5.0, 2.0),
                                (10.0, 2.0),
                                (50.0, 2.0),
                            ],
                            delay=DelayProperties(
                                enable=True, params=DelayParams(duration=0.5)
                            ),
                        ),
                        multiDispense=MultiDispenseProperties(
                            submerge=Submerge(
                                startPosition=TipPosition(
                                    positionReference=PositionReference.WELL_TOP,
                                    offset=Coordinate(x=0, y=0, z=2),
                                ),
                                speed=100,
                                delay=DelayProperties(
                                    enable=False, params=DelayParams(duration=0.0)
                                ),
                            ),
                            retract=RetractDispense(
                                endPosition=TipPosition(
                                    positionReference=PositionReference.WELL_TOP,
                                    offset=Coordinate(x=2, y=3, z=1),
                                ),
                                speed=50,
                                airGapByVolume=[(1.0, 0.1), (49.9, 0.1), (50.0, 0.0)],
                                blowout=BlowoutProperties(
                                    enable=False,
                                    params=BlowoutParams(
                                        location=BlowoutLocation.DESTINATION,
                                        flowRate=10,
                                    ),
                                ),
                                touchTip=TouchTipProperties(
                                    enable=False,
                                    params=LiquidClassTouchTipParams(
                                        zOffset=-1, mmFromEdge=0.5, speed=30
                                    ),
                                ),
                                delay=DelayProperties(
                                    enable=False, params=DelayParams(duration=0)
                                ),
                            ),
                            dispensePosition=TipPosition(
                                positionReference=PositionReference.WELL_BOTTOM,
                                offset=Coordinate(x=1, y=3, z=2),
                            ),
                            flowRateByVolume=[(50.0, 50.0)],
                            correctionByVolume=[
                                (0.0, 0.0),
                                (10.0, -1.0),
                                (50.0, -10.0),
                            ],
                            conditioningByVolume=[(1.0, 5.0), (45.0, 5.0), (50.0, 0.0)],
                            disposalByVolume=[(1.0, 5.0), (45.0, 5.0), (50.0, 0.0)],
                            delay=DelayProperties(
                                enable=True, params=DelayParams(duration=0.2)
                            ),
                        ),
                    )
                ],
            )
        ],
    )


@pytest.fixture
def minimal_transfer_properties_dict() -> Dict[str, Dict[str, TransferPropertiesDict]]:
    """A minimal dictionary representation of transfer properties of a liquid class."""
    return {
        "flex_1channel_50": {
            "opentrons/opentrons_flex_96_tiprack_50ul/1": TransferPropertiesDict(
                aspirate=AspiratePropertiesDict(
                    submerge=SubmergeDict(
                        start_position=TipPositionDict(
                            position_reference="well-bottom",
                            offset={"x": 1, "y": 2, "z": 3},
                        ),
                        speed=100,
                        delay=DelayPropertiesDict(
                            enabled=False,
                        ),
                    ),
                    retract=RetractAspirateDict(
                        end_position=TipPositionDict(
                            position_reference="well-bottom",
                            offset={"x": 1, "y": 2, "z": 3},
                        ),
                        speed=40,
                        air_gap_by_volume=[(5.0, 3.0), (10.0, 4.0)],
                        touch_tip=TouchTipPropertiesDict(enabled=False),
                        delay=DelayPropertiesDict(enabled=False),
                    ),
                    aspirate_position=TipPositionDict(
                        position_reference="well-bottom",
                        offset={"x": 1, "y": 2, "z": 3},
                    ),
                    flow_rate_by_volume=[(10.0, 40.0), (20.0, 30.0)],
                    correction_by_volume=[(0.0, 0.0)],
                    pre_wet=True,
                    mix=MixPropertiesDict(enabled=False),
                    delay=DelayPropertiesDict(enabled=False),
                ),
                dispense=SingleDispensePropertiesDict(
                    submerge=SubmergeDict(
                        start_position=TipPositionDict(
                            position_reference="well-bottom",
                            offset={"x": 1, "y": 2, "z": 3},
                        ),
                        speed=100,
                        delay=DelayPropertiesDict(
                            enabled=False,
                        ),
                    ),
                    retract=RetractDispenseDict(
                        end_position=TipPositionDict(
                            position_reference="well-bottom",
                            offset={"x": 1, "y": 2, "z": 3},
                        ),
                        speed=40,
                        air_gap_by_volume=[(5.0, 3.0), (10.0, 4.0)],
                        touch_tip=TouchTipPropertiesDict(enabled=False),
                        delay=DelayPropertiesDict(enabled=False),
                        blowout=BlowoutPropertiesDict(enabled=False),
                    ),
                    dispense_position=TipPositionDict(
                        position_reference="well-bottom",
                        offset={"x": 1, "y": 2, "z": 3},
                    ),
                    flow_rate_by_volume=[(10.0, 40.0), (20.0, 30.0)],
                    correction_by_volume=[(0.0, 0.0)],
                    push_out_by_volume=[(10.0, 7.0), (20.0, 10.0)],
                    mix=MixPropertiesDict(enabled=False),
                    delay=DelayPropertiesDict(enabled=False),
                ),
            )
        }
    }


@pytest.fixture
def custom_pip_n_tip_transfer_properties_dict() -> Dict[
    str, Dict[str, TransferPropertiesDict]
]:
    """A minimal dictionary representation of transfer properties for a custom pipette and tiprack."""
    return {
        "a_custom_pipette_type": {
            "a_custom_tiprack_uri": TransferPropertiesDict(
                aspirate=AspiratePropertiesDict(
                    submerge=SubmergeDict(
                        start_position=TipPositionDict(
                            position_reference="well-bottom",
                            offset={"x": 1, "y": 2, "z": 3},
                        ),
                        speed=100,
                        delay=DelayPropertiesDict(
                            enabled=False,
                        ),
                    ),
                    retract=RetractAspirateDict(
                        end_position=TipPositionDict(
                            position_reference="well-bottom",
                            offset={"x": 1, "y": 2, "z": 3},
                        ),
                        speed=40,
                        air_gap_by_volume=[(5.0, 3.0), (10.0, 4.0)],
                        touch_tip=TouchTipPropertiesDict(enabled=False),
                        delay=DelayPropertiesDict(enabled=False),
                    ),
                    aspirate_position=TipPositionDict(
                        position_reference="well-bottom",
                        offset={"x": 1, "y": 2, "z": 3},
                    ),
                    flow_rate_by_volume=[(10.0, 40.0), (20.0, 30.0)],
                    correction_by_volume=[(0.0, 0.0)],
                    pre_wet=True,
                    mix=MixPropertiesDict(enabled=False),
                    delay=DelayPropertiesDict(enabled=False),
                ),
                dispense=SingleDispensePropertiesDict(
                    submerge=SubmergeDict(
                        start_position=TipPositionDict(
                            position_reference="well-bottom",
                            offset={"x": 1, "y": 2, "z": 3},
                        ),
                        speed=100,
                        delay=DelayPropertiesDict(
                            enabled=False,
                        ),
                    ),
                    retract=RetractDispenseDict(
                        end_position=TipPositionDict(
                            position_reference="well-bottom",
                            offset={"x": 1, "y": 2, "z": 3},
                        ),
                        speed=40,
                        air_gap_by_volume=[(5.0, 3.0), (10.0, 4.0)],
                        touch_tip=TouchTipPropertiesDict(enabled=False),
                        delay=DelayPropertiesDict(enabled=False),
                        blowout=BlowoutPropertiesDict(enabled=False),
                    ),
                    dispense_position=TipPositionDict(
                        position_reference="well-bottom",
                        offset={"x": 1, "y": 2, "z": 3},
                    ),
                    flow_rate_by_volume=[(10.0, 40.0), (20.0, 30.0)],
                    correction_by_volume=[(0.0, 0.0)],
                    push_out_by_volume=[(10.0, 7.0), (20.0, 10.0)],
                    mix=MixPropertiesDict(enabled=False),
                    delay=DelayPropertiesDict(enabled=False),
                ),
            )
        }
    }


@pytest.fixture
def custom_pip_n_tip_transfer_properties_dict_v2() -> Dict[str, Dict[str, Any]]:
    """A minimal dictionary representation of transfer properties for a custom pipette and tiprack."""
    return {
        "a_custom_pipette_type": {
            "a_custom_tiprack_uri": {
                "aspirate": {
                    "aspirate_position": {
                        "offset": {"x": 1, "y": 2, "z": 3},
                        "position_reference": "well-bottom",
                    },
                    "correction_by_volume": [(0.0, 0.0)],
                    "delay": {"enable": False},
                    "flow_rate_by_volume": [(10.0, 40.0), (20.0, 30.0)],
                    "mix": {"enable": False},
                    "pre_wet": True,
                    "retract": {
                        "air_gap_by_volume": [(5.0, 3.0), (10.0, 4.0)],
                        "delay": {"enable": False},
                        "end_position": {
                            "offset": {"x": 1, "y": 2, "z": 3},
                            "position_reference": "well-bottom",
                        },
                        "speed": 40,
                        "touch_tip": {"enable": False},
                    },
                    "submerge": {
                        "delay": {"enable": False},
                        "speed": 100,
                        "start_position": {
                            "offset": {"x": 1, "y": 2, "z": 3},
                            "position_reference": "well-bottom",
                        },
                    },
                },
                "dispense": {
                    "dispense_position": {
                        "offset": {"x": 1, "y": 2, "z": 3},
                        "position_reference": "well-bottom",
                    },
                    "correction_by_volume": [(0.0, 0.0)],
                    "delay": {"enable": False},
                    "flow_rate_by_volume": [(10.0, 40.0), (20.0, 30.0)],
                    "mix": {"enable": False},
                    "push_out_by_volume": [(10.0, 7.0), (20.0, 10.0)],
                    "retract": {
                        "air_gap_by_volume": [(5.0, 3.0), (10.0, 4.0)],
                        "blowout": {"enable": False},
                        "delay": {"enable": False},
                        "end_position": {
                            "offset": {"x": 1, "y": 2, "z": 3},
                            "position_reference": "well-bottom",
                        },
                        "speed": 40,
                        "touch_tip": {"enable": False},
                    },
                    "submerge": {
                        "delay": {"enable": False},
                        "speed": 100,
                        "start_position": {
                            "offset": {"x": 1, "y": 2, "z": 3},
                            "position_reference": "well-bottom",
                        },
                    },
                },
                "multi_dispense": {
                    "dispense_position": {
                        "offset": {"x": 0, "y": 0, "z": 1},
                        "position_reference": "well-bottom",
                    },
                    "flow_rate_by_volume": [(0, 318)],
                    "correction_by_volume": [(0, 0)],
                    "delay": {"enabled": False},
                    "submerge": {
                        "delay": {"enabled": False},
                        "speed": 100,
                        "start_position": {
                            "offset": {"x": 0, "y": 0, "z": 2},
                            "position_reference": "well-top",
                        },
                    },
                    "retract": {
                        "air_gap_by_volume": [(0, 0)],
                        "delay": {"enabled": False},
                        "end_position": {
                            "offset": {"x": 0, "y": 0, "z": 2},
                            "position_reference": "well-top",
                        },
                        "speed": 50,
                        "touch_tip": {"enabled": False},
                        "blowout": {
                            "enabled": True,
                            "location": "trash",
                            "flow_rate": 478,
                        },
                    },
                    "conditioning_by_volume": [(0, 0)],
                    "disposal_by_volume": [(0, 5)],
                },
            }
        }
    }



import pytest

from opentrons_shared_data.labware.types import LabwareDefinition2
from opentrons_shared_data.pipette.types import PipetteNameType

from opentrons import types
from opentrons.hardware_control import ThreadManagedHardware
from opentrons.protocol_api import MAX_SUPPORTED_VERSION
from opentrons.protocol_api.core.legacy.deck import Deck
from opentrons.protocol_api.core.legacy.labware_offset_provider import (
    NullLabwareOffsetProvider,
)
from opentrons.protocol_api.core.legacy.legacy_instrument_core import (
    LegacyInstrumentCore,
)
from opentrons.protocol_api.core.legacy.legacy_labware_core import LegacyLabwareCore
from opentrons.protocol_api.core.legacy.legacy_protocol_core import (
    LegacyProtocolCore,
)
from opentrons.protocol_api.core.legacy_simulator.legacy_instrument_core import (
    LegacyInstrumentCoreSimulator,
)
from opentrons.protocol_api.core.legacy_simulator.legacy_protocol_core import (
    LegacyProtocolCoreSimulator,
)


@pytest.fixture
def protocol_context(
    hardware: ThreadManagedHardware, deck_definition_name: str
) -> LegacyProtocolCore:
    """Protocol context implementation fixture."""
    return LegacyProtocolCore(
        sync_hardware=hardware.sync,
        deck_layout=Deck(deck_type=deck_definition_name),
        api_version=MAX_SUPPORTED_VERSION,
        labware_offset_provider=NullLabwareOffsetProvider(),
    )


@pytest.fixture
def simulating_protocol_context(
    hardware: ThreadManagedHardware,
    deck_definition_name: str,
) -> LegacyProtocolCoreSimulator:
    """Protocol context simulation fixture."""
    return LegacyProtocolCoreSimulator(
        sync_hardware=hardware.sync,
        deck_layout=Deck(deck_type=deck_definition_name),
        api_version=MAX_SUPPORTED_VERSION,
        labware_offset_provider=NullLabwareOffsetProvider(),
    )


@pytest.fixture
def instrument_context(
    protocol_context: LegacyProtocolCore,
) -> LegacyInstrumentCore:
    """Instrument context backed by hardware simulator."""
    return protocol_context.load_instrument(
        PipetteNameType.P300_SINGLE_GEN2, types.Mount.RIGHT
    )


@pytest.fixture
def second_instrument_context(
    protocol_context: LegacyProtocolCore,
) -> LegacyInstrumentCore:
    """Instrument context backed by hardware simulator."""
    return protocol_context.load_instrument(
        PipetteNameType.P300_SINGLE_GEN2, types.Mount.LEFT
    )


@pytest.fixture
def simulating_instrument_context(
    simulating_protocol_context: LegacyProtocolCoreSimulator,
    instrument_context: LegacyInstrumentCore,
) -> LegacyInstrumentCoreSimulator:
    """A simulating instrument context."""
    return LegacyInstrumentCoreSimulator(
        protocol_interface=simulating_protocol_context,
        pipette_dict=instrument_context.get_hardware_state(),
        mount=types.Mount.RIGHT,
        instrument_name="p300_single_gen2",
    )


@pytest.fixture
def second_simulating_instrument_context(
    simulating_protocol_context: LegacyProtocolCoreSimulator,
    second_instrument_context: LegacyInstrumentCore,
) -> LegacyInstrumentCoreSimulator:
    """A simulating instrument context."""
    return LegacyInstrumentCoreSimulator(
        protocol_interface=simulating_protocol_context,
        pipette_dict=second_instrument_context.get_hardware_state(),
        mount=types.Mount.LEFT,
        instrument_name="p300_single_gen2",
    )


@pytest.fixture
def labware(minimal_labware_def: LabwareDefinition2) -> LegacyLabwareCore:
    """Labware fixture."""
    return LegacyLabwareCore(
        definition=minimal_labware_def,
        parent=types.Location(types.Point(0, 0, 0), "1"),
    )


@pytest.fixture
def tip_rack(minimal_labware_def: LabwareDefinition2) -> LegacyLabwareCore:
    tip_rack_definition = minimal_labware_def.copy()
    tip_rack_parameters = minimal_labware_def["parameters"].copy()

    tip_rack_parameters["isTiprack"] = True
    tip_rack_parameters["tipLength"] = 123
    tip_rack_definition["parameters"] = tip_rack_parameters
    tip_rack_definition["wells"]["A1"]["totalLiquidVolume"] = 200
    tip_rack_definition["wells"]["A2"]["totalLiquidVolume"] = 200

    """Labware fixture."""
    return LegacyLabwareCore(
        definition=tip_rack_definition,
        parent=types.Location(types.Point(0, 0, 0), "1"),
    )


@pytest.fixture
def second_labware(minimal_labware_def: LabwareDefinition2) -> LegacyLabwareCore:
    """Labware fixture."""
    return LegacyLabwareCore(
        definition=minimal_labware_def,
        parent=types.Location(types.Point(0, 0, 0), "5"),
    )



"""Data for concurrent protocol tasks."""

from datetime import datetime
from typing import TYPE_CHECKING

from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import requires_version

if TYPE_CHECKING:
    from .core.common import TaskCore


class Task:
    """A concurrent protocol task created by a protocol API function.

    *New in version 2.27*
    """

    def __init__(self, core: "TaskCore", api_version: APIVersion) -> None:
        """Initialize a Task."""
        self._core = core
        self._api_version = api_version

    @property
    @requires_version(2, 27)
    def created_at(self) -> datetime:
        """The timestamp of when the task was created."""
        return self._core.get_created_at_timestamp()

    @property
    @requires_version(2, 27)
    def done(self) -> bool:
        """Returns `True` if the task is done."""
        return self._core.is_done()

    @property
    @requires_version(2, 27)
    def started(self) -> bool:
        """Returns `True` if the task has started."""
        return self._core.is_started()
        ...

    @property
    @requires_version(2, 27)
    def finished_at(self) -> datetime | None:
        """The timestamp of the when the task finished.

        Returns `None` if the task hasn't finished yet.
        """
        return self._core.get_finished_at_timestamp()



from opentrons.hardware_control import PauseManager
from opentrons.hardware_control.types import PauseType


def test_pause_and_delay_separation() -> None:
    """
    Test that pause and delay rely on two separate pause types, and
    do not resume each other
    """
    pause_mgr = PauseManager()
    assert pause_mgr.queue == []

    pause_mgr.pause(PauseType.PAUSE)
    assert pause_mgr.queue == [PauseType.PAUSE]

    pause_mgr.pause(PauseType.DELAY)
    assert pause_mgr.queue == [PauseType.PAUSE, PauseType.DELAY]

    pause_mgr.resume(PauseType.PAUSE)
    assert pause_mgr.queue == [PauseType.DELAY]

    pause_mgr.resume(PauseType.PAUSE)
    assert pause_mgr.queue == [PauseType.DELAY]

    pause_mgr.resume(PauseType.DELAY)
    assert pause_mgr.queue == []



"""Test move to well commands."""

from datetime import datetime

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import StallOrCollisionDetectedError

from opentrons.protocol_engine import (
    DeckPoint,
    WellOffset,
)
from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.move_to_well import (
    MoveToWellImplementation,
    MoveToWellParams,
    MoveToWellResult,
)
from opentrons.protocol_engine.commands.movement_common import StallOrCollisionError
from opentrons.protocol_engine.execution import MovementHandler
from opentrons.protocol_engine.resources.model_utils import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import LabwareWellId, LiquidHandlingWellLocation
from opentrons.types import Point


@pytest.fixture
def mock_state_view(decoy: Decoy) -> StateView:
    """Get a mock StateView."""
    return decoy.mock(cls=StateView)


@pytest.fixture
def mock_model_utils(decoy: Decoy) -> ModelUtils:
    """Get a mock ModelUtils."""
    return decoy.mock(cls=ModelUtils)


async def test_move_to_well_implementation(
    decoy: Decoy,
    state_view: StateView,
    movement: MovementHandler,
    mock_model_utils: ModelUtils,
) -> None:
    """A MoveToWell command should have an execution implementation."""
    subject = MoveToWellImplementation(
        state_view=state_view, movement=movement, model_utils=mock_model_utils
    )

    data = MoveToWellParams(
        pipetteId="abc",
        labwareId="123",
        wellName="A3",
        wellLocation=LiquidHandlingWellLocation(offset=WellOffset(x=1, y=2, z=3)),
        forceDirect=True,
        minimumZHeight=4.56,
        speed=7.89,
    )

    decoy.when(
        await movement.move_to_well(
            pipette_id="abc",
            labware_id="123",
            well_name="A3",
            well_location=LiquidHandlingWellLocation(offset=WellOffset(x=1, y=2, z=3)),
            force_direct=True,
            minimum_z_height=4.56,
            speed=7.89,
            current_well=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return(Point(x=9, y=8, z=7))

    result = await subject.execute(data)

    assert result == SuccessData(
        public=MoveToWellResult(position=DeckPoint(x=9, y=8, z=7)),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="abc",
                new_location=LabwareWellId(labware_id="123", well_name="A3"),
                new_deck_point=DeckPoint(x=9, y=8, z=7),
            )
        ),
    )


async def test_move_to_well_with_tip_rack_and_volume_offset(
    decoy: Decoy,
    mock_state_view: StateView,
    movement: MovementHandler,
    mock_model_utils: ModelUtils,
) -> None:
    """It should disallow movement to a tip rack when volumeOffset is specified."""
    subject = MoveToWellImplementation(
        state_view=mock_state_view, movement=movement, model_utils=mock_model_utils
    )

    data = MoveToWellParams(
        pipetteId="abc",
        labwareId="123",
        wellName="A3",
        wellLocation=LiquidHandlingWellLocation(
            offset=WellOffset(x=1, y=2, z=3), volumeOffset=-40.0
        ),
        forceDirect=True,
        minimumZHeight=4.56,
        speed=7.89,
    )

    decoy.when(mock_state_view.labware.is_tiprack("123")).then_return(True)

    with pytest.raises(ValueError):
        await subject.execute(data)


async def test_move_to_well_stall_defined_error(
    decoy: Decoy,
    mock_state_view: StateView,
    movement: MovementHandler,
    mock_model_utils: ModelUtils,
) -> None:
    """It should catch StallOrCollisionError exceptions and make them DefinedErrors."""
    error_id = "error-id"
    error_timestamp = datetime(year=2020, month=1, day=2)
    decoy.when(
        await movement.move_to_well(
            pipette_id="abc",
            labware_id="123",
            well_name="A3",
            well_location=LiquidHandlingWellLocation(offset=WellOffset(x=1, y=2, z=3)),
            force_direct=True,
            minimum_z_height=4.56,
            speed=7.89,
            current_well=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_raise(StallOrCollisionDetectedError())
    decoy.when(mock_model_utils.generate_id()).then_return(error_id)
    decoy.when(mock_model_utils.get_timestamp()).then_return(error_timestamp)

    subject = MoveToWellImplementation(
        state_view=mock_state_view, movement=movement, model_utils=mock_model_utils
    )

    data = MoveToWellParams(
        pipetteId="abc",
        labwareId="123",
        wellName="A3",
        wellLocation=LiquidHandlingWellLocation(offset=WellOffset(x=1, y=2, z=3)),
        forceDirect=True,
        minimumZHeight=4.56,
        speed=7.89,
    )

    result = await subject.execute(data)
    assert isinstance(result, DefinedErrorData)
    assert result == DefinedErrorData(
        public=StallOrCollisionError.model_construct(
            id=error_id, createdAt=error_timestamp, wrappedErrors=[matchers.Anything()]
        ),
        state_update=update_types.StateUpdate(pipette_location=update_types.CLEAR),
    )



"""Test for the ProtocolEngine-based well API core."""

import inspect

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.labware_definition import (
    CircularWellDefinition2,
    ConicalFrustum,
    InnerWellGeometry,
    RectangularWellDefinition2,
    SphericalSegment,
    WellDefinition2,
)
from opentrons_shared_data.pipette.types import PipetteNameType

from opentrons.protocol_api import MAX_SUPPORTED_VERSION
from opentrons.protocol_api._liquid import Liquid
from opentrons.protocol_api.core.engine import WellCore, point_calculations, stringify
from opentrons.protocol_engine import (
    LoadedPipette,
    WellLocation,
    WellOffset,
    WellOrigin,
)
from opentrons.protocol_engine import commands as cmd
from opentrons.protocol_engine.clients import SyncClient as EngineClient
from opentrons.protocol_engine.errors.exceptions import (
    LiquidHeightUnknownError,
    LiquidVolumeUnknownError,
)
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import UnsupportedAPIError
from opentrons.types import Mount, MountType, Point


@pytest.fixture(autouse=True)
def patch_mock_point_calculations(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out point_calculations.py functions."""
    for name, func in inspect.getmembers(point_calculations, inspect.isfunction):
        monkeypatch.setattr(point_calculations, name, decoy.mock(func=func))


@pytest.fixture(autouse=True)
def patch_mock_stringify(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    """Mock out stringify.py functions."""
    for name, func in inspect.getmembers(stringify, inspect.isfunction):
        monkeypatch.setattr(stringify, name, decoy.mock(func=func))


@pytest.fixture
def mock_engine_client(decoy: Decoy) -> EngineClient:
    """Get a mock ProtocolEngine synchronous client."""
    return decoy.mock(cls=EngineClient)


@pytest.fixture
def api_version() -> APIVersion:
    """Get an API version to apply to the interface."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def well_definition() -> WellDefinition2:
    """Get a partial WellDefinition2 value object."""
    return CircularWellDefinition2.model_construct()  # type: ignore[call-arg]


@pytest.fixture
def subject(
    decoy: Decoy, mock_engine_client: EngineClient, well_definition: WellDefinition2
) -> WellCore:
    """Get a WellCore test subject with mocked dependencies."""
    decoy.when(
        mock_engine_client.state.labware.get_well_definition(
            labware_id="labware-id", well_name="well-name"
        )
    ).then_return(well_definition)

    return WellCore(
        name="well-name",
        labware_id="labware-id",
        engine_client=mock_engine_client,
    )


def test_name(subject: WellCore) -> None:
    """It should have a name and labware ID."""
    assert subject.get_name() == "well-name"
    assert subject.labware_id == "labware-id"


def test_display_name(
    decoy: Decoy, mock_engine_client: EngineClient, subject: WellCore
) -> None:
    """It should have a display name."""
    decoy.when(
        stringify.well(
            engine_client=mock_engine_client,
            well_name="well-name",
            labware_id="labware-id",
        )
    ).then_return("Matthew Zwimpfer")

    assert subject.get_display_name() == "Matthew Zwimpfer"


@pytest.mark.parametrize(
    "well_definition",
    [CircularWellDefinition2.model_construct(totalLiquidVolume=101)],  # type: ignore[call-arg]
)
def test_max_volume(subject: WellCore) -> None:
    """It should have a max volume."""
    assert subject.get_max_volume() == 101


def test_get_top(
    decoy: Decoy, mock_engine_client: EngineClient, subject: WellCore
) -> None:
    """It should get a well top."""
    decoy.when(
        mock_engine_client.state.geometry.get_well_position(
            labware_id="labware-id",
            well_name="well-name",
            well_location=WellLocation(
                origin=WellOrigin.TOP, offset=WellOffset(x=0, y=0, z=42)
            ),
        )
    ).then_return(Point(1, 2, 3))

    assert subject.get_top(z_offset=42.0) == Point(1, 2, 3)


def test_get_bottom(
    decoy: Decoy, mock_engine_client: EngineClient, subject: WellCore
) -> None:
    """It should get a well bottom."""
    decoy.when(
        mock_engine_client.state.geometry.get_well_position(
            labware_id="labware-id",
            well_name="well-name",
            well_location=WellLocation(
                origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=42)
            ),
        )
    ).then_return(Point(1, 2, 3))

    assert subject.get_bottom(z_offset=42.0) == Point(1, 2, 3)


def test_get_center(
    decoy: Decoy, mock_engine_client: EngineClient, subject: WellCore
) -> None:
    """It should get a well center."""
    decoy.when(
        mock_engine_client.state.geometry.get_well_position(
            labware_id="labware-id",
            well_name="well-name",
            well_location=WellLocation(origin=WellOrigin.CENTER),
        )
    ).then_return(Point(1, 2, 3))

    assert subject.get_center() == Point(1, 2, 3)


def test_has_tip(
    decoy: Decoy, mock_engine_client: EngineClient, subject: WellCore
) -> None:
    """It should get whether a clean tip is present."""
    decoy.when(
        mock_engine_client.state.tips.has_clean_tip(
            labware_id="labware-id", well_name="well-name"
        )
    ).then_return(True)

    assert subject.has_tip() is True


def test_set_has_tip(subject: WellCore) -> None:
    """Trying to set the has tip state should raise an error."""
    with pytest.raises(UnsupportedAPIError):
        subject.set_has_tip(True)


def test_get_meniscus(
    decoy: Decoy, mock_engine_client: EngineClient, subject: WellCore
) -> None:
    """Get meniscus coordinates."""
    decoy.when(
        mock_engine_client.state.geometry.get_meniscus_height(
            labware_id="labware-id",
            well_name="well-name",
        )
    ).then_return(1.23)
    decoy.when(
        mock_engine_client.state.geometry.get_well_position(
            labware_id="labware-id",
            well_name="well-name",
            well_location=WellLocation(
                origin=WellOrigin.BOTTOM,
                offset=WellOffset(x=0, y=0, z=1.23),
                volumeOffset=0.0,
            ),
        )
    ).then_return(Point(1, 2, 4.23))

    assert subject.get_meniscus() == Point(1, 2, 4.23)


def test_load_liquid(
    decoy: Decoy, mock_engine_client: EngineClient, subject: WellCore
) -> None:
    """It should load a liquid into a well."""
    mock_liquid = Liquid(
        _id="liquid-id", name="water", description=None, display_color=None
    )

    subject.load_liquid(liquid=mock_liquid, volume=20)

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.LoadLiquidParams(
                labwareId="labware-id",
                liquidId="liquid-id",
                volumeByWell={"well-name": 20},
            )
        ),
        times=1,
    )


@pytest.mark.parametrize(
    "well_definition",
    [CircularWellDefinition2.model_construct(shape="circular", diameter=123.4)],  # type: ignore[call-arg]
)
def test_diameter(subject: WellCore) -> None:
    """It should get the diameter."""
    assert subject.diameter == 123.4


@pytest.mark.parametrize(
    "well_definition",
    [RectangularWellDefinition2.model_construct(shape="rectangular", xDimension=567.8)],  # type: ignore[call-arg]
)
def test_length(subject: WellCore) -> None:
    """It should get the length."""
    assert subject.length == 567.8


@pytest.mark.parametrize(
    "well_definition",
    [RectangularWellDefinition2.model_construct(shape="rectangular", yDimension=987.6)],  # type: ignore[call-arg]
)
def test_width(subject: WellCore) -> None:
    """It should get the width."""
    assert subject.width == 987.6


@pytest.mark.parametrize(
    "well_definition",
    [CircularWellDefinition2.model_construct(depth=42.0)],  # type: ignore[call-arg]
)
def test_depth(subject: WellCore) -> None:
    """It should get the depth."""
    assert subject.depth == 42.0


def test_current_liquid_height(
    decoy: Decoy, subject: WellCore, mock_engine_client: EngineClient
) -> None:
    """Make sure current_liquid_height returns the correct meniscus value or raises an error."""
    fake_meniscus_height = 2222.2
    decoy.when(
        mock_engine_client.state.geometry.get_meniscus_height(
            labware_id="labware-id", well_name="well-name"
        )
    ).then_return(fake_meniscus_height)
    assert subject.current_liquid_height() == fake_meniscus_height

    # make sure that WellCore propagates a LiquidHeightUnknownError
    decoy.when(
        mock_engine_client.state.geometry.get_meniscus_height(
            labware_id="labware-id", well_name="well-name"
        )
    ).then_raise(LiquidHeightUnknownError())

    with pytest.raises(LiquidHeightUnknownError):
        subject.current_liquid_height()


def test_current_liquid_volume(
    decoy: Decoy, subject: WellCore, mock_engine_client: EngineClient
) -> None:
    """Make sure current_liquid_volume returns the correct value or raises an error."""
    fake_volume = 2222.2
    decoy.when(
        mock_engine_client.state.geometry.get_current_well_volume(
            labware_id="labware-id", well_name="well-name"
        )
    ).then_return(fake_volume)
    assert subject.get_liquid_volume() == fake_volume

    # make sure that WellCore propagates a LiquidVolumeUnknownError
    decoy.when(
        mock_engine_client.state.geometry.get_current_well_volume(
            labware_id="labware-id", well_name="well-name"
        )
    ).then_raise(LiquidVolumeUnknownError())

    with pytest.raises(LiquidVolumeUnknownError):
        subject.get_liquid_volume()


@pytest.mark.parametrize("operation_volume", [0.0, 100, -100, 2, -4, 5])
@pytest.mark.parametrize("mount", [Mount.LEFT, "left"])
def test_estimate_liquid_height_after_pipetting(
    decoy: Decoy,
    subject: WellCore,
    mock_engine_client: EngineClient,
    operation_volume: float,
    mount: Mount | str,
) -> None:
    """Make sure estimate_liquid_height_after_pipetting returns the correct value and does not raise an error."""
    fake_well_geometry = InnerWellGeometry(
        sections=[
            SphericalSegment(
                shape="spherical",
                radiusOfCurvature=1.0,
                topHeight=2.5,
                bottomHeight=0.0,
            ),
            ConicalFrustum(
                shape="conical",
                bottomHeight=2.5,
                topHeight=10.1,
                bottomDiameter=4.4,
                topDiameter=6.7,
            ),
            ConicalFrustum(
                shape="conical",
                bottomHeight=10.1,
                topHeight=10.2,
                bottomDiameter=6.7,
                topDiameter=7.7,
            ),
        ]
    )
    decoy.when(
        mock_engine_client.state.labware.get_well_geometry(
            labware_id="labware-id", well_name="well-name"
        )
    ).then_return(fake_well_geometry)
    initial_liquid_height = 5.6
    fake_final_height = 10000000
    decoy.when(subject.current_liquid_height()).then_return(initial_liquid_height)
    decoy.when(
        mock_engine_client.state.geometry.get_well_height_after_liquid_handling(
            labware_id="labware-id",
            well_name="well-name",
            pipette_id="pipette-id",
            initial_height=initial_liquid_height,
            volume=operation_volume,
        )
    ).then_return(fake_final_height)
    decoy.when(
        mock_engine_client.state.pipettes.get_by_mount(MountType.LEFT)
    ).then_return(
        LoadedPipette(
            id="pipette-id",
            pipetteName=PipetteNameType.P300_SINGLE,
            mount=MountType.LEFT,
        )
    )

    # make sure that no error was raised
    final_height = subject.estimate_liquid_height_after_pipetting(
        operation_volume=operation_volume, mount=mount
    )
    assert final_height == fake_final_height


def test_from_center_cartesian(
    decoy: Decoy, mock_engine_client: EngineClient, subject: WellCore
) -> None:
    """It should get the relative point from the center of a well."""
    decoy.when(
        mock_engine_client.state.geometry.get_well_position(
            labware_id="labware-id",
            well_name="well-name",
            well_location=WellLocation(origin=WellOrigin.CENTER),
        )
    ).then_return(Point(1, 2, 3))

    decoy.when(
        mock_engine_client.state.labware.get_well_size(
            labware_id="labware-id", well_name="well-name"
        )
    ).then_return((4, 5, 6))

    decoy.when(
        point_calculations.get_relative_offset(
            point=Point(1, 2, 3),
            size=(4, 5, 6),
            x_ratio=7,
            y_ratio=8,
            z_ratio=9,
        )
    ).then_return(Point(3, 2, 1))

    result = subject.from_center_cartesian(x=7, y=8, z=9)

    assert result == Point(3, 2, 1)



import typing

from opentrons.protocol_api import labware
from opentrons.protocol_api.core.legacy.legacy_labware_core import LegacyLabwareCore
from opentrons.protocols.api_support.types import APIVersion
from opentrons.types import Location, Point

if typing.TYPE_CHECKING:
    from opentrons_shared_data.labware.types import LabwareDefinition2


def test_wells_rebuilt_with_offset(minimal_labware_def: "LabwareDefinition2") -> None:
    test_labware = labware.Labware(
        core=LegacyLabwareCore(minimal_labware_def, Location(Point(0, 0, 0), "deck")),
        api_version=APIVersion(2, 13),  # set_offset() is not implemented in 2.14.
        protocol_core=None,  # type: ignore[arg-type]
        core_map=None,  # type: ignore[arg-type]
    )
    old_well_top = test_labware.wells()[0].top().point
    assert test_labware._core.get_geometry().offset == Point(10, 10, 5)  # type: ignore[attr-defined]
    assert test_labware._core.get_calibrated_offset() == Point(10, 10, 5)

    test_labware.set_offset(x=2, y=2, z=2)
    new_well = test_labware.wells()[0]
    assert new_well.top().point != old_well_top
    assert test_labware._core.get_geometry().offset == Point(10, 10, 5)  # type: ignore[attr-defined]
    assert test_labware._core.get_calibrated_offset() == Point(12, 12, 7)



from __future__ import annotations

from dataclasses import dataclass
from io import BytesIO
from typing import BinaryIO

from opentrons.protocol_reader import AbstractInputFile


@dataclass(frozen=True)
class InputFile(AbstractInputFile):
    """An implementation of AbstractInputFile to use for test input."""

    filename: str
    file: BinaryIO

    @classmethod
    def make(cls, filename: str, contents: bytes) -> InputFile:
        """Build a new InputFile with the given filename and contents."""
        return cls(filename=filename, file=BytesIO(contents))



"""Test the InstrumentContext class and its functions"""

from typing import Any, Callable, Dict
from unittest import mock

import pytest

import opentrons.protocol_api as papi
from opentrons.hardware_control import ThreadManagedHardware
from opentrons.protocols.advanced_control.transfers import transfer as v1_transfer
from opentrons.protocols.api_support.types import APIVersion
from opentrons.types import Mount

pytestmark = pytest.mark.ot2_only


@pytest.fixture
def make_context_and_labware(
    hardware: ThreadManagedHardware, deck_definition_name: str
) -> Callable[[APIVersion], Dict[str, Any]]:
    def _make_context_and_labware(api_version: APIVersion) -> Dict[str, Any]:
        ctx = papi.create_protocol_context(
            api_version=api_version,
            hardware_api=hardware,
            deck_type=deck_definition_name,
        )
        lw1 = ctx.load_labware("biorad_96_wellplate_200ul_pcr", 1)
        instr = ctx.load_instrument("p300_single", Mount.RIGHT)

        return {"ctx": ctx, "instr": instr, "lw1": lw1}

    return _make_context_and_labware


@pytest.mark.parametrize(
    "liquid_handling_command", ["transfer", "consolidate", "distribute"]
)
def test_blowout_location_unsupported_version(
    make_context_and_labware: Callable[[APIVersion], Dict[str, Any]],
    liquid_handling_command: str,
) -> None:
    # not supported in versions below 2.8
    context_and_labware = make_context_and_labware(APIVersion(2, 7))
    context_and_labware["ctx"].home()
    lw1 = context_and_labware["lw1"]
    instr = context_and_labware["instr"]
    with pytest.raises(
        ValueError,
        match="Cannot specify blowout location when using api version "
        + "below 2.8, current version is 2.7",
    ):
        getattr(instr, liquid_handling_command)(
            100, lw1["A1"], lw1["A2"], blowout_location="should not matter"
        )


@pytest.mark.parametrize(
    argnames="liquid_handling_command,blowout_location,expected_error_match,",
    argvalues=[
        ["transfer", "some invalid location", "blowout location should be either"],
        [
            "consolidate",
            "source well",
            "blowout location for consolidate cannot be source well",
        ],
        [
            "distribute",
            "destination well",
            "blowout location for distribute cannot be destination well",
        ],
    ],
)
def test_blowout_location_invalid(
    make_context_and_labware: Callable[[APIVersion], Dict[str, Any]],
    liquid_handling_command: str,
    blowout_location: str,
    expected_error_match: str,
) -> None:
    context_and_labware = make_context_and_labware(APIVersion(2, 8))
    context_and_labware["ctx"].home()
    lw1 = context_and_labware["lw1"]
    instr = context_and_labware["instr"]
    with pytest.raises(ValueError, match=expected_error_match):
        getattr(instr, liquid_handling_command)(
            100, lw1["A1"], lw1["A2"], blowout_location=blowout_location
        )


@pytest.mark.parametrize(
    argnames="liquid_handling_command,blowout_location,expected_strat,",
    argvalues=[
        ["transfer", "destination well", v1_transfer.BlowOutStrategy.DEST],
        ["transfer", "source well", v1_transfer.BlowOutStrategy.SOURCE],
        ["transfer", "trash", v1_transfer.BlowOutStrategy.TRASH],
        ["consolidate", "destination well", v1_transfer.BlowOutStrategy.DEST],
        ["consolidate", "trash", v1_transfer.BlowOutStrategy.TRASH],
        ["distribute", "source well", v1_transfer.BlowOutStrategy.SOURCE],
        ["distribute", "trash", v1_transfer.BlowOutStrategy.TRASH],
    ],
)
def test_valid_blowout_location(
    make_context_and_labware: Callable[[APIVersion], Dict[str, Any]],
    liquid_handling_command: str,
    blowout_location: str,
    expected_strat: str,
) -> None:
    context_and_labware = make_context_and_labware(APIVersion(2, 8))
    context_and_labware["ctx"].home()
    lw1 = context_and_labware["lw1"]
    instr = context_and_labware["instr"]

    with mock.patch.object(papi.InstrumentContext, "_execute_transfer") as patch:
        getattr(instr, liquid_handling_command)(
            100,
            lw1["A2"],
            [lw1["A1"], lw1["B1"]],
            blow_out=True,
            blowout_location=blowout_location,
            new_tip="never",
        )
        blowout_strat = patch.call_args[0][0]._options.transfer.blow_out_strategy

        assert blowout_strat == expected_strat



"""Tests for the InstrumentContext implementation core."""

from typing import cast

import pytest
from _pytest.fixtures import SubRequest
from decoy import Decoy

from opentrons.hardware_control import SyncHardwareAPI
from opentrons.protocol_api.core.legacy.legacy_instrument_core import (
    LegacyInstrumentCore,
)
from opentrons.protocol_api.core.legacy.legacy_protocol_core import (
    LegacyProtocolCore,
)
from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION
from opentrons.protocols.api_support.types import APIVersion
from opentrons.types import Mount


@pytest.fixture(params=[Mount.LEFT, Mount.RIGHT])
def mount(request: SubRequest) -> Mount:
    """Set the subject's mount."""
    return cast(Mount, request.param)


@pytest.fixture
def api_version() -> APIVersion:
    """Set the subject's API version."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def mock_sync_hardware_api(
    decoy: Decoy, mount: Mount, api_version: APIVersion
) -> SyncHardwareAPI:
    """Get a mocked out HardwareControlAPI dependency."""
    hw_api = decoy.mock(name="SyncHardwareAPI")
    decoy.when(hw_api.get_attached_instrument(mount)).then_return(
        {
            "default_aspirate_flow_rates": {str(api_version): 123},
            "default_dispense_flow_rates": {str(api_version): 456},
            "default_blow_out_flow_rates": {str(api_version): 789},
        }
    )
    return cast(SyncHardwareAPI, hw_api)


@pytest.fixture
def mock_protocol_impl(
    decoy: Decoy, mock_sync_hardware_api: SyncHardwareAPI
) -> LegacyProtocolCore:
    protocol_impl = decoy.mock(cls=LegacyProtocolCore)
    decoy.when(protocol_impl.get_hardware()).then_return(mock_sync_hardware_api)
    return protocol_impl


@pytest.fixture
def subject(
    mock_protocol_impl: LegacyProtocolCore,
    mount: Mount,
    api_version: APIVersion,
) -> LegacyInstrumentCore:
    return LegacyInstrumentCore(
        protocol_interface=mock_protocol_impl,
        mount=mount,
        instrument_name="cool pipette",
        default_speed=1234,
        api_version=api_version,
    )


def test_home(
    mount: Mount,
    decoy: Decoy,
    mock_sync_hardware_api: SyncHardwareAPI,
    subject: LegacyInstrumentCore,
) -> None:
    """It should home the pipette's Z-axis and plunger."""
    subject.home()

    decoy.verify(
        mock_sync_hardware_api.home_z(mount=mount, allow_home_other=True),
        mock_sync_hardware_api.home_plunger(mount),
    )


@pytest.mark.parametrize("api_version", [APIVersion(2, 12)])
def test_home_legacy_behavior(
    mount: Mount,
    decoy: Decoy,
    mock_sync_hardware_api: SyncHardwareAPI,
    subject: LegacyInstrumentCore,
) -> None:
    """It should avoid homing other instrument in API version <= 2.12."""
    subject.home()

    decoy.verify(
        mock_sync_hardware_api.home_z(mount=mount, allow_home_other=False),
        mock_sync_hardware_api.home_plunger(mount),
    )



import pytest
from mock import AsyncMock

from opentrons.drivers.asyncio.communication import AsyncSerial
from opentrons.drivers.command_builder import CommandBuilder
from opentrons.drivers.smoothie_drivers.connection import SmoothieConnection
from opentrons.drivers.smoothie_drivers.constants import SMOOTHIE_COMMAND_TERMINATOR


@pytest.fixture
def mock_serial_connection() -> AsyncMock:
    """Mock serial connection."""
    return AsyncMock(spec=AsyncSerial)


# Async because SmoothieConnection.__init__() needs an event loop,
# so this fixture needs to run in an event loop.
@pytest.fixture
async def subject(mock_serial_connection: AsyncMock) -> SmoothieConnection:
    """The test subject."""
    return SmoothieConnection(
        serial=mock_serial_connection,
        port="",
        ack="\r\n",
        name="",
        retry_wait_time_seconds=0,
        error_keyword="error",
        alarm_keyword="alarm",
    )


@pytest.fixture
def command() -> CommandBuilder:
    """A command fixture"""
    return (
        CommandBuilder(terminator=SMOOTHIE_COMMAND_TERMINATOR)
        .add_gcode("M123")
        .add_int("F", 2)
    )


async def test_command_sender_sanitized_response(
    subject: SmoothieConnection,
    mock_serial_connection: AsyncMock,
    command: CommandBuilder,
) -> None:
    """It should return sanitized result."""
    sanitized_body = "a"
    full_body = f"{command.build()}\r\n  {sanitized_body}  \r\n"

    mock_serial_connection.read_until.return_value = full_body.encode()

    result = await subject.send_command(command=command, retries=3)
    assert result == sanitized_body


@pytest.mark.parametrize(
    argnames=["cmd", "resp", "expected"],
    argvalues=[
        # Remove command from response
        ["G28.2B", "G28.2B", ""],
        ["G28.2B G1", "G28.2B G1", ""],
        ["G28.2B G1", "G1G28.2BG1", ""],
        # Remove command and whitespace from response
        ["\r\nG52\r\n\r\n", "\r\nG52\r\n\r\n", ""],
        [
            "\r\nG52\r\n\r\nsome-data\r\nok\r\n",
            "\r\nG52\r\n\r\nsome-data\r\nok\r\nTESTS-RULE",
            "TESTS-RULE",
        ],
        [
            "\r\nG52\r\n\r\nsome-data\r\nok\r\n",
            "G52\r\n\r\nsome-data\r\nokT\r\nESTS-RULE",
            "TESTS-RULE",
        ],
        # L is not a command echo but a token
        ["M371 L \r\n\r\n", "L:703130", "L:703130"],
        # R is not a command echo but a token
        ["M3 R \r\n\r\n", "M3R:703130", "R:703130"],
        [
            "M369 L \r\n\r\n",
            "M369 L \r\n\r\nL:5032304D56323032303230303432323036000000000000000000000000000000",
            "L:5032304D56323032303230303432323036000000000000000000000000000000",
        ],
    ],
)
def test_remove_serial_echo(cmd: str, resp: str, expected: str) -> None:
    """It should remove unwanted characters only."""
    res = SmoothieConnection._remove_unwanted_characters(cmd, resp)
    assert res == expected



"""Wait for duration command request, result, and implementation models."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Type

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Literal

from ..errors.error_occurrence import ErrorOccurrence
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData

if TYPE_CHECKING:
    from ..execution import RunControlHandler


WaitForDurationCommandType = Literal["waitForDuration"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class WaitForDurationParams(BaseModel):
    """Payload required to pause the protocol."""

    seconds: float = Field(..., description="Duration, in seconds, to wait for.")
    message: str | SkipJsonSchema[None] = Field(
        None,
        description="A user-facing message associated with the pause",
        json_schema_extra=_remove_default,
    )


class WaitForDurationResult(BaseModel):
    """Result data from the execution of a wait for duration command."""


class WaitForDurationImplementation(
    AbstractCommandImpl[WaitForDurationParams, SuccessData[WaitForDurationResult]]
):
    """Wait for duration command implementation."""

    def __init__(self, run_control: RunControlHandler, **kwargs: object) -> None:
        self._run_control = run_control

    async def execute(
        self, params: WaitForDurationParams
    ) -> SuccessData[WaitForDurationResult]:
        """Wait for a duration of time."""
        await self._run_control.wait_for_duration(params.seconds)
        return SuccessData(
            public=WaitForDurationResult(),
        )


class WaitForDuration(
    BaseCommand[WaitForDurationParams, WaitForDurationResult, ErrorOccurrence]
):
    """Wait for duration command model."""

    commandType: WaitForDurationCommandType = "waitForDuration"
    params: WaitForDurationParams
    result: Optional[WaitForDurationResult] = None

    _ImplementationCls: Type[WaitForDurationImplementation] = (
        WaitForDurationImplementation
    )


class WaitForDurationCreate(BaseCommandCreate[WaitForDurationParams]):
    """Wait for duration command request model."""

    commandType: WaitForDurationCommandType = "waitForDuration"
    params: WaitForDurationParams

    _CommandCls: Type[WaitForDuration] = WaitForDuration



"""Test Temperature Module's wait for temperature command implementation."""

from decoy import Decoy

from opentrons.hardware_control.modules import TempDeck
from opentrons.protocol_engine.commands import temperature_module
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.temperature_module.wait_for_temperature import (
    WaitForTemperatureImpl,
)
from opentrons.protocol_engine.execution import EquipmentHandler
from opentrons.protocol_engine.state.module_substates import (
    TemperatureModuleId,
    TemperatureModuleSubState,
)
from opentrons.protocol_engine.state.state import StateView


async def test_wait_for_temperature(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
) -> None:
    """It should be able to wait for the module's target temperature."""
    subject = WaitForTemperatureImpl(state_view=state_view, equipment=equipment)

    data = temperature_module.WaitForTemperatureParams(moduleId="tempdeck-id")

    module_substate = decoy.mock(cls=TemperatureModuleSubState)
    tempdeck_hardware = decoy.mock(cls=TempDeck)

    decoy.when(
        state_view.modules.get_temperature_module_substate(module_id="tempdeck-id")
    ).then_return(module_substate)

    decoy.when(module_substate.get_plate_target_temperature()).then_return(123)
    decoy.when(module_substate.module_id).then_return(
        TemperatureModuleId("tempdeck-id")
    )

    # Get stubbed hardware module
    decoy.when(
        equipment.get_module_hardware_api(TemperatureModuleId("tempdeck-id"))
    ).then_return(tempdeck_hardware)

    result = await subject.execute(data)
    decoy.verify(
        await tempdeck_hardware.await_temperature(awaiting_temperature=123), times=1
    )
    assert result == SuccessData(
        public=temperature_module.WaitForTemperatureResult(),
    )


async def test_wait_for_temperature_requested_celsius(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
) -> None:
    """It should be able to wait for the module's target temperature."""
    subject = WaitForTemperatureImpl(state_view=state_view, equipment=equipment)

    data = temperature_module.WaitForTemperatureParams(
        moduleId="tempdeck-id", celsius=12.3
    )

    module_substate = decoy.mock(cls=TemperatureModuleSubState)
    tempdeck_hardware = decoy.mock(cls=TempDeck)

    decoy.when(
        state_view.modules.get_temperature_module_substate(module_id="tempdeck-id")
    ).then_return(module_substate)

    decoy.when(module_substate.module_id).then_return(
        TemperatureModuleId("tempdeck-id")
    )

    # Stub temperature validation
    decoy.when(module_substate.validate_target_temperature(celsius=12.3)).then_return(
        12
    )

    # Get stubbed hardware module
    decoy.when(
        equipment.get_module_hardware_api(TemperatureModuleId("tempdeck-id"))
    ).then_return(tempdeck_hardware)

    result = await subject.execute(data)
    decoy.verify(
        await tempdeck_hardware.await_temperature(awaiting_temperature=12), times=1
    )
    assert result == SuccessData(
        public=temperature_module.WaitForTemperatureResult(),
    )



"""Tests for the ChangeNotifier interface."""

import asyncio

import pytest

from opentrons.util.change_notifier import ChangeNotifier


async def test_single_subscriber() -> None:
    """Test that a single subscriber can wait for a notification."""
    subject = ChangeNotifier()
    result = asyncio.create_task(subject.wait())

    # ensure that the wait actually waits by delaying and
    # checking that the task has not resolved
    await asyncio.sleep(0.1)
    assert result.done() is False

    asyncio.get_running_loop().call_soon(subject.notify)

    await result


@pytest.mark.parametrize("_test_repetition", range(10))
async def test_multiple_subscribers(_test_repetition: int) -> None:
    """Test that multiple subscribers can wait for a notification.

    This test checks that the subscribers are awoken in the order they
    subscribed. This may or may not be guaranteed according to the
    implementations of both ChangeNotifier and the event loop.
    This test functions as a canary, given that our code may rely
    on this ordering for determinism.

    This test runs multiple times to check for flakiness.
    """
    subject = ChangeNotifier()
    results = []

    async def _do_task_1() -> None:
        await subject.wait()
        results.append(1)

    async def _do_task_2() -> None:
        await subject.wait()
        results.append(2)

    async def _do_task_3() -> None:
        await subject.wait()
        results.append(3)

    task_1 = asyncio.create_task(_do_task_1())
    task_2 = asyncio.create_task(_do_task_2())
    task_3 = asyncio.create_task(_do_task_3())

    asyncio.get_running_loop().call_soon(subject.notify)
    await asyncio.gather(task_1, task_2, task_3)

    assert results == [1, 2, 3]


async def test_notify_while_busy() -> None:
    """Test that waiters process a new notify() after they are done being busy."""
    subject = ChangeNotifier()
    results = []

    async def _do_task() -> None:
        results.append("TEST")
        await asyncio.sleep(0.2)  # Simulate being busy

    async def do_task() -> None:
        while True:
            await subject.wait()
            await _do_task()

    task = asyncio.create_task(do_task())

    subject.notify()
    await asyncio.sleep(0.0)

    subject.notify()
    await asyncio.sleep(0.5)

    assert results == ["TEST", "TEST"]

    task.cancel()



"""Well view tests.

DEPRECATED: Testing WellView independently of WellStore is no longer helpful.
Try to add new tests to test_well_state.py, where they can be tested together,
treating WellState as a private implementation detail.
"""

from datetime import datetime

import pytest

from opentrons.protocol_engine.state.wells import WellState, WellView
from opentrons.protocol_engine.types import (
    LoadedVolumeInfo,
    ProbedHeightInfo,
    ProbedVolumeInfo,
)


@pytest.fixture
def subject() -> WellView:
    """Get a well view test subject."""
    loaded_volume_info = LoadedVolumeInfo(
        volume=30.0, last_loaded=datetime.now(), operations_since_load=0
    )
    probed_height_info = ProbedHeightInfo(height=5.5, last_probed=datetime.now())
    probed_volume_info = ProbedVolumeInfo(
        volume=25.0, last_probed=datetime.now(), operations_since_probe=0
    )
    state = WellState(
        loaded_volumes={"labware_id_1": {"well_name": loaded_volume_info}},
        probed_heights={"labware_id_2": {"well_name": probed_height_info}},
        probed_volumes={"labware_id_2": {"well_name": probed_volume_info}},
    )

    return WellView(state)


def test_get_well_liquid_info(subject: WellView) -> None:
    """Should return a tuple of well infos."""
    volume_info = subject.get_well_liquid_info(
        labware_id="labware_id_1", well_name="well_name"
    )
    assert volume_info.loaded_volume is not None
    assert volume_info.probed_height is None
    assert volume_info.probed_volume is None
    assert volume_info.loaded_volume.volume == 30.0

    volume_info = subject.get_well_liquid_info(
        labware_id="labware_id_2", well_name="well_name"
    )
    assert volume_info.loaded_volume is None
    assert volume_info.probed_height is not None
    assert volume_info.probed_volume is not None
    assert volume_info.probed_height.height == 5.5
    assert volume_info.probed_volume.volume == 25.0


def test_get_all(subject: WellView) -> None:
    """Should return a list of well summaries."""
    summaries = subject.get_all()

    assert len(summaries) == 2, f"{summaries}"
    assert summaries[0].loaded_volume == 30.0
    assert summaries[1].probed_height == 5.5
    assert summaries[1].probed_volume == 25.0



"""Tests for ProtocolEngine plugins."""

from datetime import datetime

import pytest
from decoy import Decoy

from opentrons.protocol_engine.actions import Action, ActionDispatcher, PlayAction
from opentrons.protocol_engine.plugins import AbstractPlugin, PluginStarter
from opentrons.protocol_engine.state.state import StateView


@pytest.fixture
def state_view(decoy: Decoy) -> StateView:
    """Get a mock StateView."""
    return decoy.mock(cls=StateView)


@pytest.fixture
def action_dispatcher(decoy: Decoy) -> ActionDispatcher:
    """Get a mock ActionDispatcher."""
    return decoy.mock(cls=ActionDispatcher)


class _MyPlugin(AbstractPlugin):
    def handle_action(self, action: Action) -> None:
        pass


def test_configure(
    decoy: Decoy, state_view: StateView, action_dispatcher: ActionDispatcher
) -> None:
    """The engine should be able to configure the plugin."""
    action = PlayAction(requested_at=datetime(year=2021, month=1, day=1))

    subject = _MyPlugin()
    subject._configure(
        state=state_view,
        action_dispatcher=action_dispatcher,
    )

    subject.dispatch(action)

    assert subject.state == state_view
    decoy.verify(action_dispatcher.dispatch(action))


async def test_setup_teardown(
    decoy: Decoy,
    state_view: StateView,
    action_dispatcher: ActionDispatcher,
) -> None:
    """The PluginStarter should setup and teardown plugins."""
    plugin_1 = decoy.mock(cls=AbstractPlugin)
    plugin_2 = decoy.mock(cls=AbstractPlugin)

    subject = PluginStarter(state=state_view, action_dispatcher=action_dispatcher)

    subject.start(plugin_1)
    decoy.verify(
        plugin_1._configure(state=state_view, action_dispatcher=action_dispatcher),
        action_dispatcher.add_handler(plugin_1),
        plugin_1.setup(),
    )

    subject.start(plugin_2)
    decoy.verify(
        plugin_2._configure(state=state_view, action_dispatcher=action_dispatcher),
        action_dispatcher.add_handler(plugin_2),
        plugin_2.setup(),
    )

    await subject.stop()
    decoy.verify(
        await plugin_1.teardown(),
        await plugin_2.teardown(),
    )



import typing
from copy import deepcopy
from typing import Any, Callable, Dict, List, Tuple
from unittest import mock

import pytest

from opentrons_shared_data.labware.types import LabwareDefinition2
from opentrons_shared_data.protocol.types import (
    BlowoutParams,
    DelayParams,
    FlowRateParams,
    JsonProtocolV3,
    MoveToSlotParams,
    PipetteAccessParams,
    PipetteAccessWithOffsetParams,
    StandardLiquidHandlingParams,
    TouchTipParams,
)

from opentrons.protocol_api import (
    MAX_SUPPORTED_VERSION,
    InstrumentContext,
    ProtocolContext,
    labware,
)
from opentrons.protocol_api.core.core_map import LoadedCoreMap
from opentrons.protocol_api.core.legacy.deck import Deck
from opentrons.protocol_api.core.legacy.legacy_labware_core import LegacyLabwareCore
from opentrons.protocol_api.core.legacy.legacy_well_core import LegacyWellCore
from opentrons.protocol_api.core.legacy.well_geometry import WellGeometry
from opentrons.protocol_api.core.well import AbstractWellCore
from opentrons.protocols.execution import execute
from opentrons.protocols.execution.execute_json_v3 import (
    _air_gap,
    _aspirate,
    _blowout,
    _delay,
    _dispense,
    _drop_tip,
    _get_location_with_offset,
    _get_well,
    _move_to_slot,
    _pick_up_tip,
    _set_flow_rate,
    _touch_tip,
    dispatch_json,
    load_labware_from_json_defs,
    load_pipettes_from_json,
)
from opentrons.protocols.parse import parse
from opentrons.types import Location, MountType, Point


def test_load_pipettes_from_json() -> None:
    def fake_pipette(name: str, mount: MountType) -> Tuple[str, MountType]:
        return (name, mount)

    ctx = mock.create_autospec(ProtocolContext)
    ctx.load_instrument = fake_pipette
    protocol: "JsonProtocolV3" = {
        "pipettes": {
            "aID": {"mount": "left", "name": "p10_single"},
            "bID": {"mount": "right", "name": "p50_single"},
        }
    }
    result = load_pipettes_from_json(ctx, protocol)
    assert result == {"aID": ("p10_single", "left"), "bID": ("p50_single", "right")}  # type: ignore[comparison-overlap]


def test_get_well(minimal_labware_def2: LabwareDefinition2) -> None:
    deck = Location(Point(0, 0, 0), "deck")
    mock_core = mock.create_autospec(AbstractWellCore)
    mock_map = mock.create_autospec(LoadedCoreMap)
    mock_pipette = mock.create_autospec(InstrumentContext)
    well_name = "A2"
    some_labware = labware.Labware(
        core=LegacyLabwareCore(minimal_labware_def2, deck),
        api_version=MAX_SUPPORTED_VERSION,
        protocol_core=mock_core,
        core_map=mock_map,
    )
    loaded_labware = {"someLabwareId": some_labware}
    params: PipetteAccessParams = {
        "labware": "someLabwareId",
        "well": well_name,
        "pipette": mock_pipette,
    }
    result = _get_well(loaded_labware, params)
    assert result == some_labware[well_name]


def test_set_flow_rate() -> None:
    pipette = mock.create_autospec(InstrumentContext)
    pipette.flow_rate.aspirate = 1
    pipette.flow_rate.dispense = 2
    pipette.flow_rate.blow_out = 3
    params: FlowRateParams = {"flowRate": 42}

    _set_flow_rate(pipette, params)
    assert pipette.flow_rate.aspirate == 42
    assert pipette.flow_rate.dispense == 42
    assert pipette.flow_rate.blow_out == 42


def test_load_labware_from_json_defs(
    ctx: ProtocolContext, get_labware_fixture: Callable[[str], Any]
) -> None:
    custom_trough_def = get_labware_fixture("fixture_12_trough")
    data: "JsonProtocolV3" = {
        "labwareDefinitions": {"someTroughDef": custom_trough_def},
        "labware": {
            "sourcePlateId": {
                "slot": "10",
                "definitionId": "someTroughDef",
                "displayName": "Source (Buffer)",
            },
            "destPlateId": {"slot": "11", "definitionId": "someTroughDef"},  # type: ignore[typeddict-item]
        },
    }
    loaded_labware = load_labware_from_json_defs(ctx, data)

    # objects in loaded_labware should be same objs as labware objs in the deck
    assert loaded_labware["sourcePlateId"] == ctx.loaded_labwares[10]
    # use the displayName from protocol's labware.labwareId.displayName
    assert "Source (Buffer)" in str(loaded_labware["sourcePlateId"])
    assert loaded_labware["destPlateId"] == ctx.loaded_labwares[11]
    # use the metadata.displayName from embedded def
    assert custom_trough_def["metadata"]["displayName"] in str(
        loaded_labware["destPlateId"]
    )


def test_get_location_with_offset(min_lw2: labware.Labware) -> None:
    loaded_labware = {"someLabwareId": min_lw2}
    mock_pipette = mock.create_autospec(InstrumentContext)
    params: PipetteAccessWithOffsetParams = {
        "offsetFromBottomMm": 3,
        "labware": "someLabwareId",
        "well": "A2",
        "pipette": mock_pipette,
    }
    result = _get_location_with_offset(loaded_labware, params)
    assert result == Location(Point(19, 28, 8), min_lw2["A2"])


def test_get_location_with_offset_fixed_trash(
    minimal_labware_def2: LabwareDefinition2,
) -> None:
    deck = Location(Point(0, 0, 0), "deck")
    mock_core = mock.create_autospec(AbstractWellCore)
    mock_map = mock.create_autospec(LoadedCoreMap)
    mock_pipette = mock.create_autospec(InstrumentContext)
    trash_labware_def = deepcopy(minimal_labware_def2)
    trash_labware_def["parameters"]["quirks"] = ["fixedTrash"]
    trash_labware = labware.Labware(
        core=LegacyLabwareCore(trash_labware_def, deck),
        api_version=MAX_SUPPORTED_VERSION,
        protocol_core=mock_core,
        core_map=mock_map,
    )

    loaded_labware = {"someLabwareId": trash_labware}
    params: PipetteAccessWithOffsetParams = {
        "offsetFromBottomMm": 3,
        "labware": "someLabwareId",
        "well": "A1",
        "pipette": mock_pipette,
    }

    result = _get_location_with_offset(loaded_labware, params)

    assert result == Location(Point(10, 28, 45), trash_labware["A1"])


@pytest.mark.parametrize(
    "params, expected",
    [
        ({"wait": True, "message": "m"}, [mock.call.pause(msg="m")]),
        ({"wait": 123, "message": "m"}, [mock.call.delay(seconds=123, msg="m")]),
    ],
)
def test_delay(params: DelayParams, expected: List[Any]) -> None:
    mock_context = mock.MagicMock()
    _delay(mock_context, params)

    assert mock_context.mock_calls == expected


def test_blowout() -> None:
    m = mock.MagicMock()
    m.pipette_mock = mock.create_autospec(InstrumentContext)
    m.mock_set_flow_rate = mock.MagicMock()
    mock_labware = mock.create_autospec(labware.Labware)

    params: BlowoutParams = {
        "pipette": "somePipetteId",
        "labware": "someLabwareId",
        "well": "someWell",
        "flowRate": 0,
        "offsetFromBottomMm": 0,
    }
    instruments = {"somePipetteId": m.pipette_mock}
    loaded_labware: Dict[str, labware.Labware] = {"someLabwareId": mock_labware}

    flow_rate_patch = "opentrons.protocols.execution.execute_json_v3._set_flow_rate"
    with mock.patch(flow_rate_patch, new=m.mock_set_flow_rate):
        _blowout(instruments, loaded_labware, params)

    assert m.mock_calls == [
        mock.call.mock_set_flow_rate(m.pipette_mock, params),
        mock.call.pipette_mock.blow_out(mock_labware.__getitem__()),
    ]


def test_pick_up_tip() -> None:
    pipette_mock = mock.create_autospec(InstrumentContext)
    mock_labware = mock.create_autospec(labware.Labware)
    params: PipetteAccessParams = {
        "pipette": "somePipetteId",
        "labware": "someLabwareId",
        "well": "someWell",
    }
    instruments = {"somePipetteId": pipette_mock}
    loaded_labware: Dict[str, labware.Labware] = {"someLabwareId": mock_labware}

    _pick_up_tip(instruments, loaded_labware, params)

    assert pipette_mock.mock_calls == [
        mock.call.pick_up_tip(mock_labware.__getitem__())
    ]


def test_drop_tip() -> None:
    pipette_mock = mock.create_autospec(InstrumentContext)
    mock_labware = mock.create_autospec(labware.Labware)

    params: PipetteAccessParams = {
        "pipette": "somePipetteId",
        "labware": "someLabwareId",
        "well": "someWell",
    }
    instruments = {"somePipetteId": pipette_mock}
    loaded_labware = {"someLabwareId": mock_labware}
    _drop_tip(instruments, loaded_labware, params)

    assert pipette_mock.mock_calls == [mock.call.drop_tip(mock_labware.__getitem__())]


def test_air_gap(minimal_labware_def2: LabwareDefinition2) -> None:
    m = mock.MagicMock()
    m.pipette_mock = mock.create_autospec(InstrumentContext)
    m.mock_set_flow_rate = mock.MagicMock()
    mock_core = mock.create_autospec(LegacyLabwareCore)
    mock_map = mock.create_autospec(LoadedCoreMap)

    deck = Location(Point(0, 0, 0), "deck")
    well_name = "A2"
    some_labware = labware.Labware(
        core=LegacyLabwareCore(minimal_labware_def2, deck),
        api_version=MAX_SUPPORTED_VERSION,
        protocol_core=mock_core,
        core_map=mock_map,
    )
    loaded_labware = {"someLabwareId": some_labware}
    # params = {"labware": "someLabwareId", "well": well_name}

    params: StandardLiquidHandlingParams = {
        "pipette": "somePipetteId",
        "volume": 42,
        "labware": "someLabwareId",
        "well": well_name,
        "offsetFromBottomMm": 12,
        "flowRate": 0,
    }

    instruments = {"somePipetteId": m.pipette_mock}

    loaded_labware = {"someLabwareId": some_labware}

    with mock.patch(
        "opentrons.protocols.execution.execute_json_v3._set_flow_rate",
        new=m.mock_set_flow_rate,
    ):
        _air_gap(instruments, loaded_labware, params)

    # NOTE: air_gap `height` arg is mm from well top,
    # so we expect it to equal offsetFromBottomMm - well depth.
    assert m.mock_calls == [
        mock.call.mock_set_flow_rate(m.pipette_mock, params),
        mock.call.pipette_mock.move_to(
            Location(point=Point(x=19, y=28, z=17.0), labware=some_labware[well_name])
        ),
        mock.call.pipette_mock.air_gap(42, 12 - 40),
    ]


def test_aspirate() -> None:
    m = mock.MagicMock()
    m.pipette_mock = mock.create_autospec(InstrumentContext)
    m.mock_get_location_with_offset = mock.MagicMock(
        return_value=mock.sentinel.location
    )
    m.mock_set_flow_rate = mock.MagicMock()

    params: StandardLiquidHandlingParams = {
        "pipette": "somePipetteId",
        "volume": 42,
        "labware": "someLabwareId",
        "well": "someWell",
        "flowRate": 0,
        "offsetFromBottomMm": 0,
    }
    instruments = {"somePipetteId": m.pipette_mock}

    with mock.patch(
        "opentrons.protocols.execution.execute_json_v3._get_location_with_offset",
        new=m.mock_get_location_with_offset,
    ):
        with mock.patch(
            "opentrons.protocols.execution.execute_json_v3._set_flow_rate",
            new=m.mock_set_flow_rate,
        ):
            _aspirate(instruments, mock.sentinel.loaded_labware, params)

    assert m.mock_calls == [
        mock.call.mock_get_location_with_offset(mock.sentinel.loaded_labware, params),
        mock.call.mock_set_flow_rate(m.pipette_mock, params),
        mock.call.pipette_mock.aspirate(42, mock.sentinel.location),
    ]


def test_dispense() -> None:
    m = mock.MagicMock()
    m.pipette_mock = mock.create_autospec(InstrumentContext)
    m.mock_get_location_with_offset = mock.MagicMock(
        return_value=mock.sentinel.location
    )
    m.mock_set_flow_rate = mock.MagicMock()

    params: StandardLiquidHandlingParams = {
        "pipette": "somePipetteId",
        "volume": 42,
        "labware": "someLabwareId",
        "well": "someWell",
        "flowRate": 0,
        "offsetFromBottomMm": 0,
    }
    instruments = {"somePipetteId": m.pipette_mock}

    with mock.patch(
        "opentrons.protocols.execution.execute_json_v3._get_location_with_offset",
        new=m.mock_get_location_with_offset,
    ):
        with mock.patch(
            "opentrons.protocols.execution.execute_json_v3._set_flow_rate",
            new=m.mock_set_flow_rate,
        ):
            _dispense(instruments, mock.sentinel.loaded_labware, params)

    assert m.mock_calls == [
        mock.call.mock_get_location_with_offset(mock.sentinel.loaded_labware, params),
        mock.call.mock_set_flow_rate(m.pipette_mock, params),
        mock.call.pipette_mock.dispense(42, mock.sentinel.location),
    ]


def test_touch_tip() -> None:
    location = Location(Point(1, 2, 3), "deck")
    mock_parent = mock.create_autospec(labware.Labware)
    mock_core = mock.create_autospec(LegacyLabwareCore)
    well = labware.Well(
        parent=mock_parent,
        core=LegacyWellCore(
            well_geometry=WellGeometry(
                {
                    "shape": "circular",
                    "depth": 40,
                    "totalLiquidVolume": 100,
                    "diameter": 30,
                    "x": 40,
                    "y": 50,
                    "z": 3,
                },
                parent_point=Point(10, 20, 30),
                parent_object=mock_core,
            ),
            display_name="some well",
            has_tip=False,
            name="A2",
        ),
        api_version=MAX_SUPPORTED_VERSION,
    )

    pipette_mock = mock.create_autospec(InstrumentContext, name="pipette_mock")
    mock_get_location_with_offset = mock.MagicMock(
        return_value=location, name="mock_get_location_with_offset"
    )
    mock_get_well = mock.MagicMock(return_value=well, name="mock_get_well")
    mock_set_flow_rate = mock.MagicMock(name="mock_set_flow_rate")

    params: TouchTipParams = {
        "pipette": "somePipetteId",
        "labware": "someLabwareId",
        "well": "someWell",
        "offsetFromBottomMm": 0,
    }
    instruments = {"somePipetteId": pipette_mock}

    with mock.patch(
        "opentrons.protocols.execution.execute_json_v3._get_location_with_offset",
        new=mock_get_location_with_offset,
    ):
        with mock.patch(
            "opentrons.protocols.execution.execute_json_v3._get_well", new=mock_get_well
        ):
            with mock.patch(
                "opentrons.protocols.execution.execute_json_v3._set_flow_rate",
                new=mock_set_flow_rate,
            ):
                _touch_tip(instruments, mock.sentinel.loaded_labware, params)

    # note: for this fn, order of calls doesn't matter b/c
    # we don't have stateful stuff like flow_rate
    mock_get_location_with_offset.assert_called_once_with(
        mock.sentinel.loaded_labware, params
    )
    mock_get_well.assert_called_once_with(mock.sentinel.loaded_labware, params)
    assert pipette_mock.mock_calls == [mock.call.touch_tip(well, v_offset=-70.0)]


def test_move_to_slot() -> None:
    slot_position = Location(Point(1, 2, 3), "deck")
    mock_context = mock.Mock()

    mock_deck = mock.MagicMock(spec=Deck)
    mock_deck.__contains__.return_value = True
    mock_deck.position_for.return_value = slot_position

    mock_context.deck = mock_deck
    pipette_mock = mock.create_autospec(InstrumentContext)

    instruments = {"somePipetteId": pipette_mock}

    params: MoveToSlotParams = {
        "pipette": "somePipetteId",
        "slot": "4",
        "offset": {"x": 10, "y": 11, "z": 12},
        "forceDirect": mock.sentinel.force_direct,
        "minimumZHeight": mock.sentinel.minimum_z_height,
    }

    _move_to_slot(mock_context, instruments, params)

    assert pipette_mock.mock_calls == [
        mock.call.move_to(
            Location(Point(11, 13, 15), "deck"),
            force_direct=mock.sentinel.force_direct,
            minimum_z_height=mock.sentinel.minimum_z_height,
        )
    ]


def test_dispatch_json() -> None:
    m = mock.MagicMock()
    mock_dispatcher_map = {
        "delay": m._delay,
        "blowout": m._blowout,
        "pickUpTip": m._pick_up_tip,
        "dropTip": m._drop_tip,
        "aspirate": m._aspirate,
        "dispense": m._dispense,
        "touchTip": m._touch_tip,
        "moveToSlot": m._move_to_slot,
    }

    with mock.patch(
        "opentrons.protocols.execution.execute_json_v3.dispatcher_map",
        new=mock_dispatcher_map,
    ):
        protocol_data = {
            "commands": [
                {"command": "delay", "params": "delay_params"},
                {"command": "blowout", "params": "blowout_params"},
                {"command": "pickUpTip", "params": "pickUpTip_params"},
                {"command": "dropTip", "params": "dropTip_params"},
                {"command": "aspirate", "params": "aspirate_params"},
                {"command": "dispense", "params": "dispense_params"},
                {"command": "touchTip", "params": "touchTip_params"},
                {"command": "moveToSlot", "params": "moveToSlot_params"},
            ]
        }
        context = mock.sentinel.context
        instruments = mock.sentinel.instruments
        loaded_labware = mock.sentinel.loaded_labware
        dispatch_json(context, protocol_data, instruments, loaded_labware)  # type: ignore[arg-type]

        assert m.mock_calls == [
            mock.call._delay(context, "delay_params"),
            mock.call._blowout(instruments, loaded_labware, "blowout_params"),
            mock.call._pick_up_tip(instruments, loaded_labware, "pickUpTip_params"),
            mock.call._drop_tip(instruments, loaded_labware, "dropTip_params"),
            mock.call._aspirate(instruments, loaded_labware, "aspirate_params"),
            mock.call._dispense(instruments, loaded_labware, "dispense_params"),
            mock.call._touch_tip(instruments, loaded_labware, "touchTip_params"),
            mock.call._move_to_slot(context, instruments, "moveToSlot_params"),
        ]


@typing.no_type_check
def test_dispatch_json_invalid_command() -> None:
    protocol_data = {
        "commands": [
            {"command": "no_such_command", "params": "foo"},
        ]
    }
    with pytest.raises(RuntimeError):
        dispatch_json(
            context=None,
            protocol_data=protocol_data,
            instruments=None,
            loaded_labware=None,
        )


@pytest.mark.ot2_only
def test_papi_execute_json_v3(
    monkeypatch: pytest.MonkeyPatch,
    ctx: ProtocolContext,
    get_json_protocol_fixture: Callable[[str, str, bool], str],
) -> None:
    protocol_data = get_json_protocol_fixture("3", "testAllAtomicSingleV3", False)
    protocol = parse(protocol_data, None)
    ctx.home()
    # Check that we end up executing the protocol ok
    execute.run_protocol(protocol, ctx)
